import tempfile

import pytest
from koalak.databases import JsonListDB, TxtListDB
from koalak.helpers_test import HelperTestListDatabase
from koalak.utils import temp_pathname


# ---------------- TEST ListJsonDatabase --------------- #
def test_with_helper_json_db():
    HelperTestListDatabase(JsonListDB).test()


def test_json_db_create_manual_db():
    with temp_pathname() as dbpath:
        with open(dbpath, "w") as f:
            f.write('["a", "b", 3]')
        l = JsonListDB(dbpath, autosync=True)
        assert list(l) == ["a", "b", 3]


def test_json_db_dict_instead_of_list_error():
    with temp_pathname() as dbpath:
        with open(dbpath, "w") as f:
            f.write("{}")
        with pytest.raises(TypeError):
            l = JsonListDB(dbpath, autosync=True)


# ---------------- TEST ListTxtDatabase --------------- #
def test_with_helper_txt_db():
    HelperTestListDatabase(TxtListDB, values=["a", "b", "c", "d", "e"]).test()


def test_txt_db_type():
    """Test that the DB only accept txt type"""
    with temp_pathname() as dbpath:
        l = TxtListDB(dbpath, autosync=False)
        l.append("a")
        l.sync()
        assert "a" in l

        l = TxtListDB(dbpath, autosync=False)
        assert "a" in l

        with pytest.raises(TypeError):
            l.append(123)

        with pytest.raises(TypeError):
            l.append(False)

        with pytest.raises(TypeError):
            l.insert(0, 123)

        with pytest.raises(TypeError):
            l.extend(["a", False])

        with pytest.raises(TypeError):
            l[0] = 50


def test_txt_db_no_newline():
    """Test that the DB don't accept newline"""
    with temp_pathname() as dbpath:
        l = TxtListDB(dbpath, autosync=True)
        l.append("a")
        assert "a" in l

        l = TxtListDB(dbpath, autosync=True)
        assert "a" in l

        with pytest.raises(ValueError):
            l.append("a\nb")


# TODO: test how json files are
def test_txt_db_filecontent():
    """Test how the file is inside"""
    with temp_pathname() as dbpath:
        l = TxtListDB(dbpath, autosync=True)

        l.append("a")
        assert "a" in l
        assert open(dbpath).read() == "a\n"

        l = TxtListDB(dbpath, autosync=True)
        assert open(dbpath).read() == "a\n"

        l.append("b")
        assert open(dbpath).read() == "a\nb\n"

        l.clear()
        assert open(dbpath).read() == ""

        l.append("a")
        l.append("b")
        assert open(dbpath).read() == "a\nb\n"

        l.remove("a")
        assert open(dbpath).read() == "b\n"


def test_txt_db_create_manual_db():
    with temp_pathname() as dbpath:
        with open(dbpath, "w") as f:
            f.write("a\nb\n")
        l = TxtListDB(dbpath, autosync=True)
        assert list(l) == ["a", "b"]



import koalak
import pytest
from koalak import exceptions


def test_mkhookmanager():
    framework = koalak.mkframework()

    @framework.mkhookmanager
    def hooka():
        pass

    assert framework.get_hook_manager("hooka") is hooka
    assert hooka.name == "hooka"

    @framework.mkhookmanager("hookb")
    def f():
        pass

    assert framework.get_hook_manager("hookb") is f
    assert f.name == "hookb"


def test_runall():
    framework = koalak.mkframework()

    @framework.mkhookmanager
    def binnary_opp(a, b):
        pass

    @binnary_opp.register
    def add(a, b):
        return a + b

    @binnary_opp.register
    def mult(a, b):
        return a * b

    results = binnary_opp.runall(3, 4)
    assert results == [7, 12]


def test_get_hook_manager():
    framework = koalak.mkframework()

    @framework.mkhookmanager
    def hm_a():
        pass

    @framework.mkhookmanager
    def hm_b():
        pass

    assert framework.get_hook_manager("hm_a") is hm_a
    assert framework.get_hook_manager("hm_b") is hm_b

    assert framework.get_hook_managers() == [hm_a, hm_b]


def test_get_hook():
    framework = koalak.mkframework()

    @framework.mkhookmanager
    def hm_a():
        pass

    @hm_a.register
    def f():
        pass

    @hm_a.register
    def f2():
        pass

    assert hm_a.get_hook("f") is f
    assert hm_a.get_hook("f2") is f2
    assert framework.get_hook("hm_a", "f") is f
    assert framework.get_hook("hm_a", "f2") is f2


def test_hook_already_exist():
    framework = koalak.mkframework()

    @framework.mkhookmanager
    def hm():
        pass

    @hm.register
    def f():  # noqa: F811
        pass

    with pytest.raises(exceptions.HookAlreadyExistException):

        @hm.register  # noqa: F811
        def f():  # noqa: F811
            pass


def test_hook_manager_already_exist():
    framework = koalak.mkframework()

    @framework.mkhookmanager("a")
    def a():
        pass

    with pytest.raises(exceptions.HookManagerAlreadyExistException):

        @framework.mkhookmanager("a")
        def b():
            pass



# Import decorators
from koalak.helpers.argparse_helper import ArgparseSubcmdHelper, SubcommandParser

from .bases import DirectoryDescription as D
from .bases import FileDescription as F
from .bases import frameworks, mkpluginmanager

# import databases
from .databases import (
    BaseRelationalDB,
    Database,
    DictDB,
    JsonListDB,
    ListDB,
    RelationalDB,
    TxtListDB,
    relationaldb,
)
from .decorators import add_post_init, addinit, optionalargs

get_unique_framework_name = frameworks.get_unique_framework_name
generate_unique_framework_name = frameworks.generate_unique_framework_name
get_framework = frameworks.get_framework
mkframework = frameworks.mkframework
get_frameworks = frameworks.get_frameworks


__all__ = [
    "generate_unique_framework_name",
    "get_framework",
    "mkframework",
    "get_frameworks",
    "ArgparseSubcmdHelper",
    "F",
    "D",
]



"""This module test DB interaction with framework, the DB is tested in the file test_databases"""
import koalak
import pytest
from koalak.databases import JsonListDB, TxtListDB
from koalak.utils import temp_pathname


# TODO: test db without
def test_list_db():
    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        assert len(framework.list_db) == 0

        txt_db = framework.create_list_db("txt_db", type="txt")

        assert isinstance(txt_db, TxtListDB)
        assert framework.list_db["txt_db"] is txt_db
        assert "txt_db" in framework.list_db
        assert len(framework.list_db) == 1
        assert txt_db.uri == f"{homepath}/databases/list/txt_db.txt"

        json_db = framework.create_list_db("json_db", type="json")
        assert isinstance(json_db, JsonListDB)
        assert framework.list_db["json_db"] is json_db
        assert "json_db" in framework.list_db
        assert len(framework.list_db) == 2
        assert json_db.uri == f"{homepath}/databases/list/json_db.json"


def test_list_db_with_path():
    with temp_pathname() as homepath, temp_pathname() as dbpath, temp_pathname() as dbpath2:
        framework = koalak.mkframework(homepath)
        assert len(framework.list_db) == 0

        txt_db = framework.create_list_db("txt_db", path=dbpath, type="txt")

        assert isinstance(txt_db, TxtListDB)
        assert framework.list_db["txt_db"] is txt_db
        assert "txt_db" in framework.list_db
        assert len(framework.list_db) == 1
        assert txt_db.uri == dbpath

        json_db = framework.create_list_db("json_db", path=dbpath2, type="json")
        assert isinstance(json_db, JsonListDB)
        assert framework.list_db["json_db"] is json_db
        assert "json_db" in framework.list_db
        assert len(framework.list_db) == 2
        assert json_db.uri == dbpath2


# TODO: test unamed framework with DB
def test_list_db_errors():
    # Test modifying list_db directly
    with temp_pathname() as homepath, temp_pathname() as dbpath:
        framework = koalak.mkframework(homepath)
        with pytest.raises(RuntimeError):
            framework.list_db["mydb"] = TxtListDB(dbpath)

    # test wrong type
    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        with pytest.raises(ValueError):
            db = framework.create_list_db("db", type="wrongtype")


def test_default_db_type():
    """Default db is json"""
    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        db = framework.create_list_db("db")
        assert isinstance(db, JsonListDB)



import tempfile

import pytest
from koalak.databases import JsonDictDB, TxtDictDB
from koalak.helpers_test import HelperTestDictDB
from koalak.utils import temp_pathname


# ---------------- TEST JsonDictDB --------------- #
def test_jsondictdb_with_helper():
    HelperTestDictDB(JsonDictDB).test()


def test_jsondictdb_db_create_manual_db():
    with temp_pathname() as dbpath:
        with open(dbpath, "w") as f:
            f.write('{"a": "1", "b": "2", "c":3}')
        d = JsonDictDB(dbpath, autosync=True)
        assert dict(d) == {"a": "1", "b": "2", "c": 3}


def test_jsondictdb_error_not_dict_loaded():
    with temp_pathname() as dbpath:
        with open(dbpath, "w") as f:
            f.write("[]")
        with pytest.raises(TypeError):
            JsonDictDB(dbpath, autosync=True)


# ---------------- TEST ListTxtDatabase --------------- #
def test_txtdictdb_with_helper():
    HelperTestDictDB(TxtDictDB, values=["a", "b", "c"], keys=["1", "2", "3"]).test()


def test_txtdictdb_value_type():
    """Test that the DB only accept txt type"""
    with temp_pathname() as dbpath:
        l = TxtDictDB(dbpath, autosync=False)
        l["a"] = "1"
        l.sync()
        assert "a" in l

        l = TxtDictDB(dbpath, autosync=False)
        assert "a" in l

        with pytest.raises(TypeError):
            l["a"] = 123

        with pytest.raises(TypeError):
            l["a"] = False

        with pytest.raises(TypeError):
            l.update({"b": "1", "a": False})

        # test key
        with pytest.raises(TypeError):
            l[0] = "a"

        with pytest.raises(TypeError):
            l.update({"b": "1", True: "x"})


def test_txtdictdb_no_newline():
    """Test that the DB don't accept newline"""
    with temp_pathname() as dbpath:
        l = TxtDictDB(dbpath, autosync=True)
        l["a"] = "1"
        assert "a" in l

        l = TxtDictDB(dbpath, autosync=True)
        assert "a" in l

        # test value
        with pytest.raises(ValueError):
            l["a"] = "a\nb"

        # test key
        with pytest.raises(ValueError):
            l["a\nb"] = "a"


def test_txtdictdb_sep_not_in_key():
    """Test that the DB don't accept newline"""
    with temp_pathname() as dbpath:
        l = TxtDictDB(dbpath, autosync=True)
        l["a"] = "1"
        assert "a" in l

        l = TxtDictDB(dbpath, autosync=True)
        assert "a" in l

        # test value: Seperator is allowed in value, we only split the first sep
        l["a"] = "a:b"

        # test key
        with pytest.raises(ValueError):
            l["a:b"] = "a"  # not allowed


def test_textdictdb_load_value_with_sep():
    with temp_pathname() as dbpath:
        l = TxtDictDB(dbpath, autosync=True)
        l["a"] = "1:1"
        assert dict(l) == {"a": "1:1"}

        l = TxtDictDB(dbpath, autosync=True)
        assert dict(l) == {"a": "1:1"}


# TODO: test how json files are


def test_txtdictdb_filecontent():
    """Test how the file is inside"""
    with temp_pathname() as dbpath:
        l = TxtDictDB(dbpath, autosync=True)

        l["a"] = "1"
        assert "a" in l
        assert open(dbpath).read() == "a:1\n"

        l = TxtDictDB(dbpath, autosync=True)
        assert open(dbpath).read() == "a:1\n"

        l["b"] = "2"
        assert open(dbpath).read() == "a:1\nb:2\n"

        l.clear()
        assert open(dbpath).read() == ""

        l["a"] = "1"
        l["b"] = "2"
        del l["a"]
        assert open(dbpath).read() == "b:2\n"


def test_txt_db_create_manual_db():
    with temp_pathname() as dbpath:
        with open(dbpath, "w") as f:
            f.write("a:1\n")
        l = TxtDictDB(dbpath, autosync=True)
        assert dict(l) == {"a": "1"}



from koala.decorators import addinit


def addx(self, *args, **kwargs):
    self.x = True


def addy(self, *args, **kwargs):
    self.y = True


def addz(self, *args, **kwargs):
    self.z = True


def addu(self, *args, **kwargs):
    self.u = True


def addv(self, *args, **kwargs):
    self.v = True


def test_simple():
    @addinit(addx)
    class A:
        pass

    assert A().x


def test_cls_have_init():
    @addinit(addx)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert a.a
    assert a.x
    """Brainstorm
    Pour que les deux init (decorated_init et original_init)
    soit executer c'est obligé d'executer le decorated_init
    et que le decorated_init execute l'original_init

    Comment le decorated init doit executer l'original_init?
    """


def test_cls_subclass_have_init_without_call_super():
    """If the subclass don't call"""

    @addinit(addx)
    class A:
        pass

    class B(A):
        def __init__(self):
            self.b = True

    b = B()
    assert b.b
    assert not hasattr(b, "x")


def test_cls_subclass_have_init_with_call_super():
    @addinit(addx)
    class A:
        pass

    class C(A):
        def __init__(self):
            super().__init__()
            self.c = True

    c = C()
    assert c.c
    assert c.x


def test_parent_cls_have_init_without_call_super():
    class A:
        def __init__(self):
            self.a = True

    @addinit(addx)
    class B(A):
        def __init__(self):
            self.b = True

    b = B()
    assert b.b
    assert not hasattr(b, "a")
    assert b.x


def test_parent_cls_have_init_with_call_super():
    class A:
        def __init__(self):
            self.a = True

    @addinit(addx)
    class B(A):
        def __init__(self):
            super().__init__()
            self.b = True

    b = B()
    assert b.b
    assert b.a
    assert b.x


"""
n: new_init (decorator)
o: old_init (__init__)

A: n(o)
B: n(o)

decorate init A
    inits exist pas crée inits
    rajoute __init__ au inits
    remplacer l'init
decorate B
    inits exist rajouter l'in
m2: chaque class on lui ajoute son init

Si class à pas d'init

An Ao
Bn Bo(s)
Bn => Bo => An

Rajouter un traitement X à l'initialisation d'un objet sans rien chambouler
def special_init(self, *args, **kwargs):
    self.x = "caca"

@addinit(special_init)
class A:
   def __init__(self):
       ...

X: spécial init (genre vérifie que la class est abstraite)
D: dispatcher (le nouveau __init__ qui sait qui appeler)
I: Init original

@D(X) A - I
@D(Y) B - I(S)
B()
D(B) c'est lui qui va mener la danse
D(B) => Y
D(B) => I  le I appel
I(S) == D(A) => (le B appele le super qui est le disptacher de A)
    appel X et appel Y

forcément n(B) va etre appelé
- execute n(b)
- n(b) va executer old (jusque la c'est bien)
- old va executer super (bien mais il va rappeler)

Je dois savoir si j'ai déjà executer init ou pas sur l'objet
- marquer l'objet!

Problem le même disptacher est appeler, deux solutions possibles
- faire en sorte qu'un dispatcher différent soit appelé
- le dispatcher gere tout

le même dispatcher est appeler veut dire

- un dispatccher qui gere tout =>

@A: x init
@B: y init (no super)
B(): b y (pas de x a)

@A: x init
@B: y init

"""


def test_two_subclasses_have_decorator_with_super():
    @addinit(addx)
    class A:
        def __init__(self):
            self.a = True

    @addinit(addy)
    class B(A):
        def __init__(self):
            super().__init__()
            self.b = True

    # Test that B don't affect A
    a = A()
    assert a.a
    assert a.x
    assert not hasattr(a, "b")
    assert not hasattr(a, "y")

    b = B()
    assert b.a
    assert b.b
    assert b.x
    assert b.y


def test_three_subclasses_with_super():
    @addinit(addx)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert a.a
    assert not hasattr(a, "b")
    assert not hasattr(a, "c")
    assert a.x
    assert not hasattr(a, "y")
    assert not hasattr(a, "z")

    @addinit(addy)
    class B(A):
        def __init__(self):
            super().__init__()
            self.b = True

    a = A()
    assert a.a
    assert not hasattr(a, "b")
    assert not hasattr(a, "c")
    assert a.x
    assert not hasattr(a, "y")
    assert not hasattr(a, "z")

    b = B()
    assert b.a
    assert b.b
    assert not hasattr(b, "c")
    assert b.x
    assert b.y
    assert not hasattr(b, "z")

    @addinit(addz)
    class C(B):
        def __init__(self):
            super().__init__()
            self.c = True

    a = A()
    assert a.a
    assert not hasattr(a, "b")
    assert not hasattr(a, "c")
    assert a.x
    assert not hasattr(a, "y")
    assert not hasattr(a, "z")

    b = B()
    assert b.a
    assert b.b
    assert not hasattr(b, "c")
    assert b.x
    assert b.y
    assert not hasattr(b, "z")

    c = C()
    assert c.a
    assert c.b
    assert c.c
    assert c.x
    assert c.y
    assert c.z


def test_args():
    @addinit(addx)
    class A:
        def __init__(self, a):
            self.a = a

    a = A(5)
    assert a.x
    assert a.a == 5


def test_args_kwargs():
    def myinit_x(self, a, b=2):
        self.x = a + b

    @addinit(myinit_x)
    class A:
        def __init__(self, a, b=2):
            self.a = a
            self.b = b

    a1 = A(5)
    assert a1.a == 5
    assert a1.b == 2
    assert a1.x == 7

    a2 = A(2, 3)
    assert a2.a == 2
    assert a2.b == 3
    assert a2.x == 5


def test_runfirst():
    def init_10(self):
        self.a = 10

    @addinit(init_10, runfirst=False)
    class A:
        def __init__(self):
            self.a = 5

    a = A()
    assert a.a == 10

    @addinit(init_10, runfirst=True)
    class A:
        def __init__(self):
            self.a = 5

    a = A()
    assert a.a == 5


def test_runfirst_default():
    # Testt that by default runfirst if False
    # ie: the added__init__ will run at the end
    # to not break code of others
    def init_10(self):
        self.a = 10

    @addinit(init_10, runfirst=False)
    class A:
        def __init__(self):
            self.a = 5

    a = A()
    assert a.a == 10


def test_many_added__inits__in_addinit():
    @addinit(addx, addy)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert hasattr(a, "x")
    assert hasattr(a, "y")
    assert hasattr(a, "a")


def test_many_addinit_in_same_cls():
    @addinit(addy)
    @addinit(addx)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert hasattr(a, "x")
    assert hasattr(a, "y")
    assert hasattr(a, "a")


def test_many_addinit_and_many_addes_inits():
    @addinit(addz, addu)
    @addinit(addx, addy)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert hasattr(a, "x")
    assert hasattr(a, "y")
    assert hasattr(a, "z")
    assert hasattr(a, "u")
    assert hasattr(a, "a")


def test_parent_without_init():
    @addinit(addx)
    class A:
        pass

    class B(A):
        def __init__(self):
            super().__init__()
            self.b = True

    b = B()
    assert b.b
    assert b.x


def test_complecated_01():
    @addinit(addx, addy)
    class A:
        def __init__(self):
            self.a = True

    a = A()
    assert a.a
    assert a.x
    assert a.y

    @addinit(addz, addu)
    @addinit(addv)
    class B(A):
        def __init__(self):
            super().__init__()
            self.b = True

    a = A()
    assert a.a
    assert a.x
    assert a.y
    assert not hasattr(a, "b")
    assert not hasattr(a, "z")
    assert not hasattr(a, "u")
    assert not hasattr(a, "v")

    b = B()
    assert b.a
    assert b.x
    assert b.y

    assert b.b
    assert b.z
    assert b.u
    assert b.v



class KoalaException(Exception):
    pass


class FrameworkAlreadyExistException(KoalaException):
    pass


class PluginManagerAlreadyExistException(KoalaException):
    pass


class PluginAlreadyExistException(KoalaException):
    pass


class HookManagerAlreadyExistException(KoalaException):
    pass


class HookAlreadyExistException(KoalaException):
    pass


class VariableAlreadyExistException(KoalaException):
    pass



import io
import os
import random
import shutil
import string
import sys
import tempfile
import types
from contextlib import contextmanager
from typing import Any, Dict, Iterable


@contextmanager
def tmp_module(
    module_name: str = None,
    context: Dict[str, Any] = None,
    addsys: bool = True,
    overwrite=False,
):
    """Generate a temporal module as context.
    Temporally add the module to sys.modules so it can be importable
    Args:
    """
    # TODO: unit test this function
    if not module_name:
        module_name = "tmpmodule_" + randomstr(exclude=list(sys.modules.keys()))
    else:
        if module_name in sys.modules and not overwrite:
            raise TabError("Module name already exist, call it with overwrite = True")

    context = context or {}
    module = types.ModuleType(module_name)
    module.__dict__.update(context)
    sys.modules[module_name] = module
    yield module
    del sys.modules[module_name]


@contextmanager
def temp_pathname():
    """Return an available pathname that you can use for files/dirs

    Will automatically clean the pathname if it's
    a file or directory"""
    tmpdir = tempfile.gettempdir()
    pathname = os.path.join(
        tmpdir, randomstr(20, prefix="tmppathname_", exclude=os.listdir(tmpdir))
    )
    try:
        yield pathname
    finally:
        if os.path.isdir(pathname):
            shutil.rmtree(pathname)
        elif os.path.isfile(pathname):
            os.remove(pathname)


def randomstr(
    n: int = 10,
    *,
    prefix="",
    suffix="",
    alphabet: str = string.ascii_letters,
    exclude: Iterable = None,
):
    """Generate random string of length n
    Args:
        n: length of returned string (without the prefix)
        alphabet: alphabet to use
        exclude: generated string must not be in this list (or any iterable object)
    """
    if exclude is not None:
        exclude = set(exclude)
    else:
        exclude = set()

    while True:
        string = prefix + "".join(random.choices(alphabet, k=n)) + suffix
        if string not in exclude:
            return string


@contextmanager
def temp_str2filename(string: str) -> str:
    """Generate a temporary file with the string content and
    return it's filename
    """
    with temp_pathname() as pathname:
        with open(pathname, "w") as f:
            f.write(string)
        yield pathname


@contextmanager
def str2stdin(string: str):
    string_io = io.StringIO(string)
    old_stdin = sys.stdin
    sys.stdin = string_io
    yield
    sys.stdin = old_stdin


@contextmanager
def file2stdin(filename: str):
    with open(filename) as f:
        old_stdin = sys.stdin
        sys.stdin = f
        try:
            yield
        # FIXME unit test finally with all other contextmanager
        finally:
            sys.stdin = old_stdin


def get_prefixed_callables_of_object(obj, prefix: str):
    """Get all the methods with a specific prefix of an object
    Example:
        get_methods_with_prefix(obj, "test_")
    """
    return [
        getattr(obj, e)
        for e in dir(obj)
        if e.startswith(prefix) and callable(getattr(obj, e))
    ]


def humaintime(time: float) -> str:
    """Return the given bytes as a human friendly KB, MB, GB, or TB string."""
    if time < 60:
        label = "seconds"
    elif time < 3600:  # 60 ** 2
        label = "minutes"
        time /= 60
    elif time < 216000:  # 60 ** 3
        label = "hours"
        time /= 3600
    else:
        label = "days"
        time /= 86400
    return f"{time:.2f} {label}"


def humanbytes(size: int) -> str:
    """Return the given bytes as a human friendly KB, MB, GB, or TB string."""
    if size < 1024:
        label = "Bytes"
    elif size < 1_048_576:  # 1024 ** 2
        label = "KB"
        size /= 1024
    elif size < 1_073_741_824:  # 1024 ** 3
        label = "MB"
        size /= 1_048_576
    else:
        label = "GB"
        size /= 1_073_741_824

    if label == "Bytes":
        return f"{size} Bytes"
    else:
        return f"{size:.2f} {label}"



"""THIS CODE IS A WORK IN PROGRESS"""
import inspect

fixtures = {}


def register_fixture(f):
    global fixtures
    fixtures[f.__name__] = f
    return f


@register_fixture
def one():
    return 1


@register_fixture
def two():
    return 2


print(fixtures)


def callfixture(f):
    params = inspect.signature(f).parameters
    args = []
    for param_name, param in params.items():
        if param_name in fixtures:
            args.append(fixtures[param_name])
        else:
            raise ValueError(f"fixture {param_name} unknown")

    def new_f():
        n_args = [arg() for arg in args]
        return f(*n_args)

    return new_f


@callfixture
def hey(one, two):
    return one + two


@register_fixture
def extra():
    return "Whaaaat"


@callfixture
def f(extra, one):
    print(extra)
    print(one)
    return extra * 2


print(f())
print(hey())



from .databases import (
    Database,
    DictDB,
    JsonDictDB,
    JsonListDB,
    ListDB,
    TxtDictDB,
    TxtListDB,
)
from .relationaldb import BaseRelationalDB
from .relationaldb_sqlalchemy import RelationalDB, relationaldb



import os

from koalak.utils import get_prefixed_callables_of_object, randomstr, temp_pathname


def test_randomstr():
    # By default the function work (no need the first param)
    randomstr()

    # test length
    for _ in range(20):
        assert len(randomstr(10)) == 10
        assert len(randomstr(20)) == 20

    # test alphabet
    alphabet = "abc"
    for _ in range(20):
        a = randomstr(alphabet=alphabet)
        for c in a:
            assert c in alphabet

    # test exclude
    alphabet = "ab"
    a = randomstr(1, alphabet=alphabet, exclude="a")
    assert a == "b"

    # test prefix
    for _ in range(20):
        assert randomstr(prefix="ab_").startswith("ab_")


def test_temp_pathname():
    # Test that if we create a file, it is correctly removed
    with temp_pathname() as pathname:
        saved_pathname = pathname
        assert not os.path.exists(pathname)
        # create the file
        open(pathname, "w")

        assert os.path.exists(pathname)
    assert not os.path.exists(saved_pathname)

    # Test that if we create a directory, it is correctly removed
    with temp_pathname() as pathname:
        saved_pathname = pathname
        assert not os.path.exists(pathname)
        # create the file
        os.makedirs(os.path.join(pathname, "test"))
        assert os.path.exists(pathname)
    assert not os.path.exists(saved_pathname)

    # Test that the file is correctly removed after an exception
    class DummyException(Exception):
        pass

    try:
        with temp_pathname() as pathname:
            # create the file
            open(pathname, "w")
            raise DummyException
    except DummyException:
        assert not os.path.exists(pathname)


def test_get_prefixed_callables_of_object():
    class A:
        test_nop = "something"

        def f(self):
            pass

        def test_a(self):
            pass

        def test_b(self):
            pass

    a = A()
    assert get_prefixed_callables_of_object(a, "test_") == [a.test_a, a.test_b]

    class B:
        test_nop = "something"
        run_z = "something"

        def f(self):
            pass

        def test_a(self):
            pass

        def test_b(self):
            pass

        def run_x(self):
            pass

        def run_y(self):
            pass

    b = B()
    assert get_prefixed_callables_of_object(b, "test_") == [b.test_a, b.test_b]
    assert get_prefixed_callables_of_object(b, "run_") == [b.run_x, b.run_y]



import os
import tempfile

import pytest
from koalak import F, generate_unique_framework_name, mkframework
from koalak.frameworks.home_structure import BashrcFileAction


def test_bashrc_empty_bashrc():
    with tempfile.TemporaryDirectory() as dirname, tempfile.NamedTemporaryFile(
        "w"
    ) as tmpfile:
        # fmt: off
        home = [F("script.sh", content="whoami", actions=BashrcFileAction(tmpfile.name))]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        name = generate_unique_framework_name()
        framework = mkframework(name, homepath=home_path, home_structure=home)
        framework.init()

        script_path = os.path.join(home_path, "script.sh")
        expected = BashrcFileAction._add_to_bashrc_template.format(
            path=script_path, basename="script.sh", framework_name=name
        )

        with open(tmpfile.name) as f:
            data = f.read()
            assert data == expected
        assert open(script_path).read() == "whoami"


def test_bashrc_not_empty_bashrc():
    """Test action bashtc with an empty bashrc"""
    with tempfile.TemporaryDirectory() as dirname, tempfile.NamedTemporaryFile(
        "w"
    ) as tmpfile:
        bashrc = "alias ls='ls --color=auto'"
        tmpfile.file.write(bashrc)
        tmpfile.file.flush()

        # fmt: off
        home = [F("script.sh", content="whoami", actions=BashrcFileAction(tmpfile.name))]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        name = generate_unique_framework_name()
        framework = mkframework(name, homepath=home_path, home_structure=home)
        framework.init()

        script_path = os.path.join(home_path, "script.sh")
        expected = bashrc
        expected += BashrcFileAction._add_to_bashrc_template.format(
            path=script_path, basename="script.sh", framework_name=name
        )

        with open(tmpfile.name) as f:
            data = f.read()
            assert data == expected
        assert open(script_path).read() == "whoami"


@pytest.mark.skip
def test_bashrc_uninstall():
    with tempfile.TemporaryDirectory() as dirname, tempfile.NamedTemporaryFile(
        "w"
    ) as tmpfile:
        # init bashrc file
        bashrc = "alias ls='ls --color=auto'"
        tmpfile.file.write(bashrc)
        tmpfile.file.flush()

        # fmt: off
        home = [F("script.sh", content="whoami", actions=BashrcFileAction(tmpfile.name))]
        # fmt: on
        home_path = os.path.join(dirname, "home")
        name = generate_unique_framework_name()
        framework = mkframework(name, homepath=home_path, home_structure=home)
        framework.init()

        script_path = os.path.join(home_path, "script.sh")
        expected = bashrc
        expected += BashrcFileAction._add_to_bashrc_template.format(
            path=script_path, basename="script.sh", framework_name=name
        )

        # check that bashrc added the needed lines
        with open(tmpfile.name) as f:
            data = f.read()
            assert data == expected
        # check the content of our script
        assert open(script_path).read() == "whoami"

        # check that when uninstalling the framework the bashrc return to it's original
        framework.uninstall()
        with open(tmpfile.name) as f:
            data = f.read()
            assert data == bashrc



import argparse
import shlex

import pytest
from koalak import ArgparseSubcmdHelper


def test_run_cmd(capsys):
    class SimpleCommand(ArgparseSubcmdHelper):
        def parser_mycmd(self, parser):
            pass

        def run_mycmd(self, args):
            print(1)

    simple_command = SimpleCommand()

    # assert that "mycmd" parse
    simple_command.parse_args(["mycmd"])

    capsys.readouterr()  # ignore the first prints
    simple_command.run(["mycmd"])
    captured = capsys.readouterr()
    assert captured.out == "1\n"


def test_run_with_args(capsys):
    class SimpleCommand(ArgparseSubcmdHelper):
        def parser_mycmd(self, parser):
            parser.add_argument("printme")

        def run_mycmd(self, args):
            print(args.printme)

    simple_command = SimpleCommand()

    # assert that "mycmd" parse
    simple_command.parse_args(["mycmd", "1"])

    with pytest.raises(SystemExit):
        simple_command.parse_args(["mycmd"])

    capsys.readouterr()  # ignore the first prints

    simple_command.run(["mycmd", "test"])
    captured = capsys.readouterr()
    assert captured.out == "test\n"

    simple_command.run(["mycmd", "22"])
    captured = capsys.readouterr()
    assert captured.out == "22\n"


def test_run_without_parse():
    """run_x without parse_x should not be a problem"""

    class SimpleCommand(ArgparseSubcmdHelper):
        def run_mycmd(self, args):
            pass

    simple_command = SimpleCommand()
    simple_command.parse_args(["mycmd"])
    simple_command.run(["mycmd"])


def test_parse_without_run():
    """parse_x without run_x should not be a problem"""

    class SimpleCommand(ArgparseSubcmdHelper):
        def parser_mycmd(self, args):
            pass

    # parser_mycmd exist but run_mycmd not present
    with pytest.raises(RuntimeError):
        SimpleCommand()


def test_run_without_subcmd():
    class SimpleCommand(ArgparseSubcmdHelper):
        def run_mycmd(self, args):
            pass

    simple_command = SimpleCommand()
    with pytest.raises(SystemExit):
        simple_command.run([])  # run without any inputs


def test_run_non_existing_cmd():
    class SimpleCommand(ArgparseSubcmdHelper):
        name = "simple"

        def run_mycmd(self, args):
            pass

    simple_command = SimpleCommand()
    with pytest.raises(SystemExit):
        simple_command.run(["dontexist"])


def test_run_attribute_not_callable():
    class SimpleCommand(ArgparseSubcmdHelper):
        run_x = "String and not callable"

    with pytest.raises(TypeError):
        SimpleCommand()


def test_nested_commands(capsys):
    class VarCommand(ArgparseSubcmdHelper):
        def run_ls(self, args):
            print("ls")

        def run_add(self, args):
            print("add")

    class ToolsmanagerCommand(ArgparseSubcmdHelper):
        command_var = VarCommand

        def run_version(self, args):
            print("version")

    toolsmanager_command = ToolsmanagerCommand()

    capsys.readouterr()  # ignore the first prints
    toolsmanager_command.run(["version"])
    captured = capsys.readouterr()
    assert captured.out == "version\n"

    toolsmanager_command.run(["var", "ls"])
    captured = capsys.readouterr()
    assert captured.out == "ls\n"

    toolsmanager_command.run(["var", "add"])
    captured = capsys.readouterr()
    assert captured.out == "add\n"


def test_nested_commands_depth_2(capsys):
    class Depth3Command(ArgparseSubcmdHelper):
        def run_a(self, args):
            print("d3.a")

        def run_b(self, args):
            print("d3.b")

    class Depth2Command(ArgparseSubcmdHelper):
        command_depth = Depth3Command

        def run_a(self, args):
            print("d2.a")

        def run_b(self, args):
            print("d2.b")

    class Depth1Command(ArgparseSubcmdHelper):
        command_depth = Depth2Command

        def run_a(self, args):
            print("d1.a")

        def run_b(self, args):
            print("d1.b")

    command = Depth1Command()

    capsys.readouterr()  # ignore the first prints
    command.run(["a"])
    captured = capsys.readouterr()
    assert captured.out == "d1.a\n"

    capsys.readouterr()  # ignore the first prints
    command.run(["b"])
    captured = capsys.readouterr()
    assert captured.out == "d1.b\n"

    with pytest.raises(SystemExit):
        command.run([])

    # depth2
    capsys.readouterr()  # ignore the first prints
    command.run(["depth", "a"])
    captured = capsys.readouterr()
    assert captured.out == "d2.a\n"

    capsys.readouterr()  # ignore the first prints
    command.run(["depth", "b"])
    captured = capsys.readouterr()
    assert captured.out == "d2.b\n"

    with pytest.raises(SystemExit):
        command.run(["depth"])

    # depth3
    capsys.readouterr()  # ignore the first prints
    command.run(["depth", "depth", "a"])
    captured = capsys.readouterr()
    assert captured.out == "d3.a\n"

    capsys.readouterr()  # ignore the first prints
    command.run(["depth", "depth", "b"])
    captured = capsys.readouterr()
    assert captured.out == "d3.b\n"

    with pytest.raises(SystemExit):
        command.run(["depth", "depth"])

    # None existing commands/args
    with pytest.raises(SystemExit):
        command.run(["a", "dontexist"])

    with pytest.raises(SystemExit):
        command.run(["depth", "dontexist"])

    with pytest.raises(SystemExit):
        command.run(["depth", "depth", "dontexist"])

    with pytest.raises(SystemExit):
        command.run(["a", "-x"])


def test_run_x_and_command_x_error():
    class VarCommand(ArgparseSubcmdHelper):
        pass

    class ToolsmanagerCommand(ArgparseSubcmdHelper):
        command_var = VarCommand

        def run_var(self, args):
            pass

    with pytest.raises(RuntimeError):
        ToolsmanagerCommand()


def test_error_command_x_wrong_type():
    class Command(ArgparseSubcmdHelper):
        # command_x must be a class ArgparseSUbcmdHelper
        command_x = "hello"

    with pytest.raises(TypeError):
        Command()


"""
Tests todo
==========
- more testing with args and run them
- generated help (groups)

"""



"""THIS CODE IS A WORK IN PROGRESS"""
import collections
import inspect
import logging
import typing
from datetime import datetime

import attr
import coloring
import sqlalchemy
from koalak.consts import PACKAGE_NAME
from koalak.decorators import add_post_init, optionalargs
from sqlalchemy import (
    Boolean,
    Column,
    Date,
    DateTime,
    ForeignKey,
    Integer,
    String,
    create_engine,
)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker

logger = logging.getLogger("koalak")
logger.setLevel(logging.DEBUG)

try:
    typping_ForwardRef = typing.ForwardRef
except AttributeError:
    typping_ForwardRef = typing._ForwardRef  # noqa


class ProgrammingError(Exception):
    """Things that should not happen!"""

    pass


"""
RelationalDB is a wrapper for sqlalchemy and attr to have
a relational database with a better API and abstraction
- no ID
- specify relations with annoation (List["Animal"], Animal, ...)
- use attr to create __init__, __repr__, ...
See examples in test_relationaldb.py



"""

"""
TODO: Error tests
- Default None -> Nullable True
- Not null and default is None => error
-
"""
AttribType = type(attr.ib())
AttrNOTHING = attr.ib()._default


def parse_typing_type(t):
    if len(t.__args__) != 1:
        raise ValueError("Annotation list must have one and only one argument")

    return {"type": t._name, "name": t.__args__[0].__forward_arg__}


@attr.s
class ColumnMetadata:
    """Helper cls to build relational attribute"""

    name: str = attr.ib()
    type = attr.ib(kw_only=True, default=None)
    many: bool = attr.ib(default=None)
    referenced_by: "ColumnMetadata" = attr.ib(default=None)
    attr_attribute = attr.ib(default=None)
    unique = attr.ib(default=None)
    nullable = attr.ib(default=None)
    default = attr.ib(default=None)

    # this attribute is set to True by the class that contain this
    #  attribute, sins it is possible for an other class to create
    #  an attribute with ref (having exist=True we are sure that
    #  the attribute exist and it's not just a reference)
    exist = attr.ib(default=None)

    def update(
        self,
        *,
        many=None,
        type=None,
        referenced_by=None,
        attr_attribute=None,
        exist=None,
        unique=None,
        nullable=None,
        default=None,
    ):
        for varname in [
            "exist",
            "many",
            "type",
            "unique",
            "referenced_by",
            "attr_attribute",
            "nullable",
            "default",
        ]:

            var = locals()[varname]
            if var is not None:
                attr = getattr(self, varname)
                if attr is not None:
                    # attr not None and var not None! they must be the same
                    if var is not attr:
                        raise ValueError(
                            f"Trying to update {varname} but it's already set to {attr} "
                        )
                else:
                    setattr(self, varname, var)

    def print(self):
        print_str = f"\t{self.name}: "
        if self.many:
            print_str += "List["
        print_str += f"{self.type}"
        if self.many:
            print_str += "]"

        if self.unique:
            print_str += " [unique]"
        if self.nullable:
            print_str += " [nullable]"
        if self.referenced_by:
            print_str += f" [{self.referenced_by}]"
        if self.exist:
            print_str += " [exists]"
        coloring.print_bblue_violet(print_str)


@attr.s
class TableMetadata:
    """Helper class to postpone building the ORM classes
    Contains all information about a table (name, indexes, ..., ...)"""

    # cls_name is the only information that we always know
    #   either we have it before decorating with @db.table
    #   or it's name is referenced as an annotation in a column
    cls_name: str = attr.ib()
    table_name: str = attr.ib(default=None)  # tablename
    cls = attr.ib(default=None, init=False)
    columns: typing.Dict[str, ColumnMetadata] = attr.ib(factory=dict, init=False)
    orm_cls = attr.ib(default=None, init=False)
    # this attribute is set to True by the class that contain this
    #  attribute, sins it is possible for an other class to create
    #  an attribute with ref (having exist=True we are sure that
    #  the attribute exist and it's not just a reference)
    exist = attr.ib(default=None)

    def update(self, *, cls=None, table_name=None, orm_cls=None, exist=None):
        for varname in ["cls", "table_name", "orm_cls", "exist"]:
            var = locals()[varname]
            if var is not None:
                attr = getattr(self, varname)
                if attr is not None:
                    # attr not None and var not None! they must be the same
                    if var is not attr:
                        raise ValueError(
                            f"Trying to update {varname} but it's already set to {attr} "
                        )
                else:
                    setattr(self, varname, var)

    def print(self):
        coloring.print_bblue(f"Table: {self.table_name} - {self.cls_name}")
        for column in self.columns.values():
            column.print()

    def get_or_create_column(self, attribute_name: str) -> ColumnMetadata:
        if attribute_name in self.columns:
            return self.columns[attribute_name]
        else:
            attribute = ColumnMetadata(attribute_name)
            self.columns[attribute_name] = attribute
            return attribute


class RelationalDB:
    # TODO: add a new engine which is dict in contrast to SQLalchemy
    # it will have the same API, but it will store the classes
    # in memory with list and dicts and don't use SQL
    # TODO: add bijection argument (for friends, to add the same line twice)
    # TODO: add List of strings List[str] => and add a table with a foreign key
    #   Ex: tags: List[str]  -> create table users_tags id_tag
    # TODO: use only annotation without db.attribute(): Ex: name: str  (like dataclass)
    # TODO: add the possibility ti use @db.table  and take the name cls.__name__
    # TODO: implement  lazy import
    _map_cls_to_sqlalchemy_types = {
        str: sqlalchemy.String,
        int: sqlalchemy.Integer,
        bool: sqlalchemy.Boolean,
        datetime: sqlalchemy.DateTime,
    }
    List = typing.List

    def __init__(self, uri: str = None, echo=None, autocommit=None):
        """
        RelationalDB is an aditional layer of abstration for relationaldb ORMs
        This class is based on SQLAlchemy and attrs modules

        Algorithm:
            - Gather information for tables (table) and columns (attributes)
              - add hook in init for each table/class to be able to autocommit
            - call init() to built SQLAlchemy classes based on the gathered information

        Global constraints:
            - Mapped classes (with tables) must have unique names
        """
        if uri is None:
            uri = ":memory:"

        if echo is None:
            echo = False

        if autocommit is None:
            autocommit = True

        self.uri = uri
        self.autocommit = autocommit
        self.debug = echo

        # internal attributes
        # FIXME: remove one of the three mapping
        self._map_cls_tablename = {}
        self._map_tablename_cls = {}

        # sqlalchemy attributes
        self.engine = create_engine(f"sqlite:///{uri}", echo=echo)
        self.Session = sessionmaker(bind=self.engine)
        self.Base = declarative_base()

        # dictionary to hold classes to build at init()
        self._clsname_to_table_metadata: typing.Dict[str, TableMetadata] = {}

    @property
    def query(self):
        return self.session.query

    @property
    def commit(self):
        return self.session.commit

    datetime_now = datetime.utcnow

    def metadata(self, unique=None, nullable=None, target=None, ref=None):
        if unique is None:
            unique = False

        if nullable is None:
            nullable = False

        return {
            "koalak": {
                "unique": unique,
                "nullable": nullable,
                "target": target,
                "ref": ref,
            }
        }

    def attribute(
        self,
        # attrs attribute
        type=None,
        *,
        default=attr.NOTHING,
        validator=None,
        repr=True,
        cmp=None,
        hash=None,
        init=True,
        metadata=None,
        converter=None,
        factory=None,
        kw_only=False,
        eq=None,
        order=None,
        on_setattr=None,
        # sqlalchemy attribute
        unique=None,
        nullable=None,
        target=None,
        ref=None,
    ):

        # check integrity of arguments

        if not nullable and default is None:
            assert TypeError(
                f"An attribute can't be 'not nullable' and default is None in the same time"
            )

        # if default is None activate nullable to True by default
        if default is None and nullable is None:
            nullable = True

        if metadata is None:
            metadata = {}
        if nullable is None:
            nullable = False
        metadata.update(
            self.metadata(unique=unique, target=target, nullable=nullable, ref=ref)
        )
        attrib = attr.ib(  # noqa
            default=default,
            validator=validator,
            repr=repr,
            cmp=cmp,
            hash=hash,
            init=init,
            metadata=metadata,
            type=type,
            converter=converter,
            factory=factory,
            kw_only=kw_only,
            eq=eq,
            order=order,
            on_setattr=on_setattr,
        )
        self.log(f"\tParsing attrib {attrib}")
        coloring.print_red("metadata", attrib.metadata)
        return attrib

    def log(self, *args, **kwargs):
        if self.debug:
            print("[DEBUG]", *args, **kwargs)

    def table(self, table_name):
        """Create new table

        Algorithm:
        This will only add metadata to create the table at init()
        - Decorate the cls with attr module to create the used methods (__init__, __repr__, ...)
        - Hook the __init__ for autocommit (so the class will add itself to session)
        - collect all needed information for relations (to add in init())
        - Create the sqlalchemy ORM cls that inherit from the original cls and the Base cls
         without the relations (relations will be added in init())
        - Return the ORM cls
        """
        self.log(f"Creating table {table_name}")
        self.print()

        def decorator(cls):
            """Register this class to build it when init() is called"""
            # This code is executed after attr.ib and db.attribute constructions
            # check if cls name is unique
            cls_name = cls.__name__
            table_metadata = self._get_or_create_table_metadata_by_clsname(cls_name)
            table_metadata.update(table_name=table_name, cls=cls)

            # TODO: check if cls exists
            # if cls_name in self._clsname_to_table_metadata:
            #    assert KeyError(f"Table with cls name {cls_name} already exists")

            # ======================================= #
            # decorate cls with attr (if not already) #
            # ======================================= #
            if not hasattr(cls, "__attrs_attrs__"):
                # If columns are created with attr.ib and not db.attribute we add koalak metadata
                # We have to create it here, before attr.s parse the attr.ib instances
                for e in cls.__dict__.values():
                    if not isinstance(e, AttribType):
                        continue
                    if "koalak" not in e.metadata:
                        e.metadata["koalak"] = self.metadata()["koalak"]
                    # FIXME: check if attr.ib and db.attribute are coherant

                # Apply attr
                cls = attr.s(cls)
            else:
                raise ValueError(
                    f"Class {cls.__name__} must not be decorated with attr.s use db.table instead!"
                )

            # ========================================== #
            # hook __init__ to add autocommit if enabled #
            # ========================================== #
            def postinit_autocommit(self_instance):
                if self.autocommit:
                    self.session.add(self_instance)
                    # self.session.commit()

            # add postinit_autocommit to the cls
            cls = add_post_init(postinit_autocommit)(cls)

            # register cls and it's attributes
            # orm_cls = self._create_orm_cls_without_relations(cls, table_name)
            self._add_table_metadata(cls, table_name)
            # building_cls = self._register_cls(cls, table_name)

            return cls

        return decorator

    def print(self):
        print("<----------BEGIN-------->")
        for table in self._clsname_to_table_metadata.values():
            table.print()
            print()
        print("<----------END-------->")

    def _get_or_create_table_metadata_by_clsname(self, cls_name: str) -> TableMetadata:
        if cls_name in self._clsname_to_table_metadata:
            return self._clsname_to_table_metadata[cls_name]
        else:
            table_metadata = TableMetadata(cls_name)
            self._clsname_to_table_metadata[cls_name] = table_metadata
            return table_metadata

    def _add_table_metadata(self, cls, table_name):
        """Create the ORM cls without relations. Relations will be added when init() is called"""

        # get the building_class for this class
        cls_name = cls.__name__
        table_metadata = self._get_or_create_table_metadata_by_clsname(cls_name)
        table_metadata.update(exist=True)

        columns_metadata = table_metadata.columns
        attr_attributes = getattr(cls, "__attrs_attrs__")
        """sqlalchemy_attributes = {
            "__tablename__": table_name,
            # Automatically add id
            "id": Column(Integer, primary_key=True),
        }"""

        for attr_attribute in attr_attributes:
            metadata = attr_attribute.metadata["koalak"]
            if attr_attribute.default is None:
                pass
            coloring.print_green("attr", attr_attribute)
            attribute_name = attr_attribute.name

            column_metadata = table_metadata.get_or_create_column(attribute_name)

            column_metadata.update(
                unique=metadata["unique"], nullable=metadata["nullable"]
            )
            # Get sqlalchemy column type
            attr_type = attr_attribute.type
            """
            We can have the following types:
            - Simple types (builtin)
                - SQLAlchemy cls -> do nothing
                - Known builtin type (int, float, bool, str, datetime) -> convert to SQLAlchemy cls
            - Complicated types (relationship with other tables)
                - Cls already present in our tables (cls already decorated with @db.table)
                - Str -> convert it to cls in our table
                - List["str"] or List[cls], many relation with the cls
            - Else error!
            """
            self.print()

            if isinstance(attr_type, type(sqlalchemy.String)):
                # sqlalchemy type (Integer, String, ...)
                column_metadata.update(type=attr_type, many=False)
            elif attr_type in self._map_cls_to_sqlalchemy_types:
                # convert builtin types  (int, str, bool, ...)
                sqlalchemy_type = self._map_cls_to_sqlalchemy_types[attr_attribute.type]
                column_metadata.update(type=attr_type, many=False)
            elif isinstance(attr_type, (str, type(typing.List))) or (
                isinstance(attr_type, type)
                and attr_type.__name__ in self._clsname_to_table_metadata
            ):
                # Complicated type implying relations (one to many, many to many, one to one, ...)
                self._add_relational_column(table_metadata, attr_attribute)

                # check it cls type is the same as this cls
                # if so, nullable must be True! (otherwise it's impossible to instantiate the object)
                # TODO: move the check in _register_building_attribute?
                """if clsname_type == cls.__name__:
                    if not metadata["nullable"]:
                        # TODO: test this case (tell Yassmine?)
                        raise ValueError(
                            f"Attribute {attr_attribute.name!r} must be nullable when "
                            f"it's type is referencing to the same class ({cls.__name__!r})"
                        )
                continue"""
            else:
                raise TypeError(f"Unknown type for attribute {attr_type}")

            # table_metadata.get_or_create_column("")
            # Build kwargs
            # kwargs["unique"] = metadata["unique"]
            # kwargs["nullable"] = metadata["nullable"]
            # Create the Column
            # sqlalchemy_attributes[attribute_name] = Column(
            #     sqlalchemy_type, *args, **kwargs
        #  )

        # sqlalchemy_attributes.update(sqlalchemy_attributes_later)
        # ORM cls inherit from the decorated cls and self.Base
        # OrmCls = type(cls.__name__, (cls, self.Base), sqlalchemy_attributes)
        # ============================================================= #
        # Create a new cls that inherit from Base and the decorated cls #
        # ============================================================= #
        # self._map_tablename_cls[table_name] = OrmCls
        # self._map_cls_tablename[OrmCls] = table_name

        # table_metadata.update(orm_cls=OrmCls)
        # return OrmCls

    def _add_relational_column(
        self, table_metadata: TableMetadata, attr_attribute
    ) -> ColumnMetadata:
        """Register a relational attribute (str or isinstance(type(List)) or cls"""
        coloring.print_green(
            "Adding relational attribute",
            repr(table_metadata.table_name),
            attr_attribute,
        )
        column_name = attr_attribute.name
        attr_type = attr_attribute.type

        # If it's a known Column class, get its name
        if (
            isinstance(attr_type, type)
            and attr_type.__name__ in self._clsname_to_table_metadata
        ):
            attr_type = attr_type.__name__

        metadata = attr_attribute.metadata["koalak"]

        cls = table_metadata.cls
        cls_name = cls.__name__
        column_metadata = table_metadata.get_or_create_column(column_name)
        column_metadata.update(attr_attribute=attr_attribute)
        column_metadata.update(exist=True)
        # check if the attribute is typing.List to use a "many" relationship

        many = False
        if isinstance(attr_type, type(typing.List)):
            parsed_type = parse_typing_type(attr_type)
            attr_type_type = parsed_type["type"]
            if attr_type_type != "List":
                raise ValueError(
                    f"Support only List annotation, don't support {attr_type}"
                )

            many = True
            attr_type = parsed_type["name"]  # noqa

        atomic_type = attr_type
        print("whoop", column_metadata)
        column_metadata.print()
        column_metadata.update(many=many, type=attr_type)
        column_metadata.print()
        # build the ref_type if it doesn't exist
        ref_cls_name = atomic_type
        ref_building_cls = self._get_or_create_table_metadata_by_clsname(atomic_type)

        # check if the attribute have a reference to an other one
        ref_attribute_name = metadata["ref"]
        if ref_attribute_name:
            # check: if the ref_class is already parsed check if ref attribute exist
            if ref_building_cls.exist:
                # if ref_attribute_name not in
                if (
                    ref_attribute_name not in ref_building_cls.columns
                    or not ref_building_cls.columns[ref_attribute_name].exist
                ):
                    raise ValueError(
                        f"{table_metadata.table_name}.{column_name} reference non existing attribute {ref_attribute_name}"
                    )

            # get the building_attribute
            ref_building_attributes = ref_building_cls.columns
            building_attribute_ref = ref_building_cls.get_or_create_column(
                ref_attribute_name
            )

            building_attribute_ref.update(type=cls_name)
            building_attribute_ref.update(referenced_by=column_metadata)

            # Check if reference are consistent
            if building_attribute_ref.referenced_by:
                # check that class is consistent
                #  ex: If Person have an attributes List[Animal] animals, ref=owner
                #  owner must be of type Person!
                ref_attomic_type = building_attribute_ref.type
                if ref_attomic_type and ref_attomic_type != cls_name:
                    raise ValueError(
                        f"{cls_name}.{column_name} reference {ref_cls_name}.{ref_attribute_name}, but "
                        f" {ref_cls_name}.{ref_attribute_name} reference {ref_attomic_type} class"
                    )
                ref_referenced_by_name = building_attribute_ref.referenced_by.name

                if ref_referenced_by_name != column_name:
                    raise ValueError(
                        f"The {cls_name}.{column_name} reference {ref_cls_name}.{ref_attribute_name}"
                        f" but {ref_cls_name}.{ref_attribute_name} "
                        f"reference {cls_name}.{ref_referenced_by_name}"
                    )

        return column_metadata

    def init(self):

        self.print()
        # Check classes!
        for table_metadata in self._clsname_to_table_metadata.values():
            for column_metadata in table_metadata.columns.values():
                for attribute_name in ["many", "unique", "nullable", "type"]:
                    attribute_value = getattr(column_metadata, attribute_name)
                    if attribute_value is None:
                        raise ProgrammingError(
                            f"In column {table_metadata.table_name}.{column_metadata.name}"
                            f" attribute {attribute_name!r} is None"
                        )

        # construict classes
        for cls_name in self._clsname_to_table_metadata:
            self._build_orm_relations(cls_name)

        self.Base.metadata.create_all(self.engine)
        self.session = self.Session()

    def _build_orm_relations(self, cls_name: str):
        buildingn_cls = self._clsname_to_table_metadata[cls_name]
        cls = buildingn_cls.cls
        orm_cls = buildingn_cls.orm_cls
        for building_attribute in buildingn_cls.columns.values():
            # Check if the attribute exist or if it's referenced but don't exist
            if not building_attribute.exist:
                referenced_by = building_attribute.referenced_by
                assert ValueError(
                    f"Attribute {building_attribute.name} is referenced by {referenced_by.name} but don't exist"
                )
            attr_attribute = building_attribute.attr_attribute
            metadata = attr_attribute.metadata["koalak"]
            ref = metadata["ref"]
            many = building_attribute.many
            if not ref:
                ref_many = True
            else:
                ref_many = True

            if not isinstance(building_attribute.type, str):
                ref_type = building_attribute.type._name
            else:
                ref_type = building_attribute.type
            ref_building_cls = self._clsname_to_table_metadata[ref_type]
            if not many and ref_many:
                # one to many relations
                self._build_one_to_many_relation(
                    buildingn_cls, ref_building_cls, building_attribute.name
                )

    def _build_one_to_many_relation(
        self,
        one_cls: TableMetadata,
        many_cls: TableMetadata,
        one_attribute_name,
        one_relation=True,
    ):
        # add the foreign_key in the one_cls side

        many_cls_id = f"{many_cls.table_name}.id"

        foreign_key = Column(Integer, ForeignKey(many_cls_id))
        foreign_key_name = f"{one_attribute_name}_id"
        setattr(one_cls.orm_cls, foreign_key_name, foreign_key)

        if one_relation:

            relation = sqlalchemy.orm.relationship(
                many_cls.name,
                remote_side=many_cls.orm_cls.id,
                foreign_keys=f"{one_cls.name}.{foreign_key_name}",
            )
            setattr(one_cls.orm_cls, one_attribute_name, relation)

    def first(self, table, **kwargs):
        if isinstance(table, str):
            table = self._map_tablename_cls[table]
        return self.session.query(table).filter_by(**kwargs).first()

    def close(self):
        self.commit()

    # aliases
    ib = attribute

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.commit()


def relationaldb(
    uri: str = None, *, debug: bool = None, autocommit: bool = None
) -> RelationalDB:
    return RelationalDB(uri, echo=debug, autocommit=autocommit)



import functools
import logging

import pytest
from koalak.decorators import optionalargs

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def test_optionalargs_without_brackets():
    @optionalargs
    def add_attr(obj, attr="x", value=5):
        setattr(obj, attr, value)
        return obj

    @add_attr(attr="y", value=10)
    class A:
        pass

    assert A.y == 10

    @add_attr
    class A:
        pass

    assert A.x == 5

    assert add_attr.__name__ == "add_attr"


def test_optionalargs_with_brackets():
    @optionalargs()
    def add_attr(obj, attr="x", value=5):
        setattr(obj, attr, value)
        return obj

    @add_attr(attr="y", value=10)
    class A:
        pass

    assert A.y == 10

    @add_attr
    class A:
        pass

    assert A.x == 5

    assert add_attr.__name__ == "add_attr"


def test_optionalargs_firtsarg_is_str():
    @optionalargs(firstarg=str)
    def add_name_attr(obj, name="name"):
        obj.name = name
        return obj

    logger.debug("Before without brackets")

    @add_name_attr
    class A:
        pass

    assert A.name == "name"

    logger.debug("Before str as argument")

    @add_name_attr("test")
    class A:
        pass

    assert A.name == "test"


def test_optionalargs_firtsarg_is_str_two_positional_args():
    @optionalargs(firstarg=str)
    def add_attr(obj, attr="x", value=5):
        setattr(obj, attr, value)
        return obj

    logger.debug("Before without brackets")

    @add_attr
    class A:
        pass

    assert A.x == 5

    logger.debug("Before str as argument")

    @add_attr("y")
    class A:
        pass

    assert A.y == 5

    @add_attr("z", 10)
    class A:
        pass

    assert A.z == 10


@pytest.mark.skip(reason="Not implemented yet")
def test_defaultdecorator_decorated():
    # @defaultdecorator(decorated=types.FunctionType)
    def add_name_attr(obj, name="name"):
        obj.name = name
        return obj

    assert add_name_attr.__name__ == "add_name_attr"

    @add_name_attr
    def f():
        pass

    assert f.name == "name"

    @add_name_attr(name="yes")
    def f():
        pass

    assert f.name == "yes"

    # positional argument work because the decorated object
    # is a function type
    @add_name_attr("yy")
    def f():
        pass

    assert f.name == "yy"

    # kwargs always works
    @add_name_attr(name="y")
    class A:
        pass

    assert A.name == "y"

    with pytest.raises(Exception):
        # positional argument on class should not work
        # because we expect types.FunctionType
        @add_name_attr
        class A:
            pass

        assert A.name == "nn"


def test_optionalargs_method_in_cls():
    class Decorator:
        @optionalargs()
        def add_attr(self, obj, attr="x", value=5):
            setattr(obj, attr, value)
            return obj

    d = Decorator()

    @d.add_attr()
    class A:
        pass

    assert A.x == 5

    @d.add_attr
    class B:
        pass

    assert B.x == 5

    @d.add_attr(attr="y")
    class C:
        pass

    assert C.y == 5

    @d.add_attr(value=10)
    class D:
        pass

    assert D.x == 10


def test_optionalargs_method_in_cls_with_firstarg():
    class Decorator:
        @optionalargs(firstarg=str)
        def add_attr(self, obj, attr="x", value=5):
            setattr(obj, attr, value)
            return obj

    d = Decorator()

    @d.add_attr()
    class A:
        pass

    assert A.x == 5

    @d.add_attr
    class B:
        pass

    assert B.x == 5

    @d.add_attr(attr="y")
    class C:
        pass

    assert C.y == 5

    @d.add_attr(value=10)
    class D:
        pass

    assert D.x == 10

    @d.add_attr("z", value=10)
    class D:
        pass

    assert D.z == 10


def test_optionalargs_method_in_cls_with_firstarg_when_cls_have_methods():
    # This is a test to avoid retrograding (I had the bug)
    class Decorator:
        # when having __len__ return 0 object will evaluate to false
        # and will bug if we test without "is not None"
        def __len__(self):
            return 0

        @optionalargs(firstarg=str)
        def add_attr(self, obj, attr="x", value=5):
            setattr(obj, attr, value)
            return obj

    d = Decorator()

    @d.add_attr()
    class A:
        pass

    assert A.x == 5

    @d.add_attr
    class B:
        pass

    assert B.x == 5

    @d.add_attr(attr="y")
    class C:
        pass

    assert C.y == 5

    @d.add_attr(value=10)
    class D:
        pass

    assert D.x == 10

    @d.add_attr("z", value=10)
    class D:
        pass

    assert D.z == 10


def test_optionalargs_modifying_function():
    @optionalargs
    def add_x_to_result(func, x=1):
        @functools.wraps(func)
        def _func(*args, **kwargs):
            result = func(*args, **kwargs)
            return result + x

        return _func

    # without brackets
    @add_x_to_result
    def ret_1():
        return 1

    assert ret_1() == 2

    # with brackets and no args
    @add_x_to_result()
    def ret_1():
        return 1

    assert ret_1() == 2

    # with brackets and optional arg
    @add_x_to_result(x=4)
    def ret_1():
        return 1

    assert ret_1() == 5



import collections
import glob
import inspect
import os
import shutil
import typing
from string import Template
from typing import Any, Callable, Dict, List, Union

import coloring
import toml
from koalak.frameworks import home_structure
from koalak.frameworks.home_structure import (
    DirectoryDescription,
    FileDescription,
    normalize_home_structure,
)

from . import databases, exceptions
from .consts import HOMEPATH_NAME, PACKAGE_NAME

"""This module contain the main bases for the project koalak (framework, plugin managers, ...)"""


def __init_subclass__(subcls, **kwargs):

    if "name" not in subcls.__dict__:
        raise ValueError(f"Base plugin must have the attribute 'name'")

    # add the plugin
    name = subcls.__dict__["name"]
    pm = subcls._fwm["plugin_manager"]
    if name in pm._plugins:
        raise exceptions.PluginAlreadyExistException(f"plugin {name} already exist")

    pm._plugins[name] = subcls

    # abstract class
    if FwmConfig.is_abstract(subcls):
        pass


def init_subclass(subcls, **kwargs):
    subcls._pluginmanager._register(subcls)


class FwmConfig:
    """Class responsible to handle internal config in objects
    like plugins. Ex: @abstract"""

    dict_name = PACKAGE_NAME

    @classmethod
    def update(cls, key, value):
        pass

    @classmethod
    def get(cls, key):
        pass

    @classmethod
    def set_abstract(cls, target_cls):
        cls.update("abstract", True)

    @classmethod
    def is_abstract(cls, target_cls):
        pass


class KoalakContainer:
    """Main container for koalak classes (framework, plugin_manager, hooks, ...)
    Act between dict and list, store named item in dict and unamed item in list
    """

    def __init__(self):
        self._dict_data = {}
        self._list_data = []

    def _add_item(self, object, name=None):
        # FIXME: item must be a string?
        if name is None:
            self._list_data.append(object)
        else:
            self._dict_data[name] = object

    def __contains__(self, item):
        if isinstance(item, str):
            return item in self._dict_data
        else:
            return item.name in self._dict_data

    def __getitem__(self, item):
        return self._dict_data[item]

    def __iter__(self):
        yield from self._dict_data.values()
        yield from self._list_data

    def __len__(self):
        return len(self._dict_data) + len(self._list_data)


class Plugin:
    pass


def mkpluginmanager(*args, **kwargs):
    return PluginManager(*args, **kwargs)


class ConstraintAttr:
    def __init__(
        self, name: str = None, type=None, inheritable: bool = True, choices=None
    ):
        """

        Args:
            name: name of the class attribute
            type: type of the class attribute
            inheritable: if the attribute can be inherited (True)
                or if each class must define the class attribute (False)
        """
        self.name = name
        self.type = type
        self.inheritable = inheritable
        self.choices = choices

    def check(self, plugin):
        if self.name is None:
            raise TypeError("Must set the attribute name")
        if self.inheritable:
            has_attr = hasattr(plugin, self.name)
        else:
            has_attr = self.name in plugin.__dict__

        if not has_attr:
            raise TypeError(
                f"plugin {plugin.__name__!r} must have the class attribute {self.name!r}"
            )

        plugin_attr = getattr(plugin, self.name)
        # Check type
        if self.type is not None:
            if not isinstance(plugin_attr, self.type):
                raise TypeError(
                    f"class attribute {self.name} must be of type {self.type}"
                )

        # Check choices
        if self.choices:
            if plugin_attr not in self.choices:
                raise ValueError(
                    f"class attribute {self.name} must be in {self.choices!r}"
                )


class PluginManager(KoalakContainer):
    """

    How it work:

    """

    def __init__(self, name: str = None, *, homepath=None):
        super().__init__()

        # FIXME: plugin without name?
        self.name: str = name
        self.baseplugin = None
        self.homepath = homepath
        self._plugins = {}
        self._initialized = False
        self._constraints = []

    def __repr__(self):
        if self.name:
            return f"<PluginManager [{self.name}]>"
        else:
            return f"<PluginManager>"

    def __str__(self):
        return self.__repr__()

    def __call__(self, **kwargs):
        for plugin in self:
            if not hasattr(plugin, "metadata"):
                continue

            skip_iteration = False
            for key, value in kwargs.items():
                print(key, plugin.metadata.get(key), value)
                if plugin.metadata.get(key) != value:
                    skip_iteration = True
                    continue
            if skip_iteration:
                continue
            yield plugin

    def _assert_baseplugin(self):
        if self.baseplugin is None:
            raise TypeError("BasePlugin not initiated!")

    def _register(self, plugin):
        if self.baseplugin is None:
            raise TypeError(
                "Can't register plugin without baseplugin. use mkbaseplugin before"
            )

        # Fixme: better think abstract class
        is_abstract = "abstract" in plugin.__dict__
        if not is_abstract:
            # Check constraint before registring
            for contraint in self._constraints:
                contraint.check(plugin)

            self._add_item(plugin, plugin.name)

    def mkbaseplugin(self, baseplugin):
        self.baseplugin = baseplugin
        baseplugin._pluginmanager = self
        baseplugin.__init_subclass__ = classmethod(init_subclass)

        attrs_to_del = []  # we can't del in loop
        for name, constraint in baseplugin.__dict__.items():
            if isinstance(constraint, ConstraintAttr):
                constraint.name = name
                self._constraints.append(constraint)
                attrs_to_del.append(name)
        for attr_name in attrs_to_del:
            delattr(baseplugin, attr_name)

        # All custom plugins (loaded from home) have _is_home_plugin to True
        self.baseplugin._is_home_plugin = False
        return baseplugin

    def init(self, _homepath_initialized: set = None):
        _homepath_initialized = _homepath_initialized or set()
        if self.homepath is None:
            raise TypeError("You can not init a plugin mananger without homepath")

        if self._initialized:
            raise TypeError("Plugin Already initiated")
        self._initialized = True

        if self.homepath in _homepath_initialized:
            return
        self._init_home()
        self._load_plugins()

    def attr(self, type=None, inheritable=True, choices=None):
        return ConstraintAttr(type=type, inheritable=inheritable, choices=choices)

    def get_home_plugins(self):
        """Get plugins loaded from home/plugins"""
        for e in self:
            if e._is_home_plugin:
                yield e

    def _load_plugins(self):
        """Load home plugins"""
        for python_path in glob.glob(os.path.join(self.homepath, "*.py")):
            with open(python_path) as f:
                data = f.read()
                execution_context = {}
                exec(data, execution_context)
                for object_name, object in execution_context.items():
                    if inspect.isclass(object) and issubclass(object, self.baseplugin):
                        if object is self.baseplugin:
                            continue
                        object._is_home_plugin = True

    def _init_home(self):
        if not self.homepath:
            return
        if not os.path.exists(self.homepath):
            os.makedirs(self.homepath)
        elif os.path.isfile(self.homepath):
            raise NotADirectoryError("Home path is not a directory")
        # else it's a directory already created


class HookManager:
    def __init__(self, name, spec_func):
        self.name = name
        self.spec_func = spec_func
        self._hooks = {}

    def register(self, func):
        func_name = func.__name__
        if func_name in self._hooks:
            raise exceptions.HookAlreadyExistException(
                f"hook {func_name} already exist"
            )

        self._hooks[func_name] = func
        return func

    def get_hook(self, hook_name: str) -> Callable:
        return self._hooks[hook_name]

    def get_hooks(self) -> List[Callable]:
        return list(self._hooks.values())

    def runall(self, *args, **kwargs):
        results = []
        for func in self._hooks.values():
            result = func(*args, **kwargs)
            results.append(result)
        return results


class Printer:
    pass


NOT_ALLOWED_VARIABLE_NAMES = ["home", "userhome"]


class FrameworkVariables(collections.UserDict):
    def __setitem__(self, key, value):
        if key in NOT_ALLOWED_VARIABLE_NAMES:
            raise TypeError(f"Variable name {key!r} is not allowed")
        super().__setitem__(key, value)

    # FIXME: what if key is not a string?
    def set(self, key: str, value: Any, substitute=False) -> None:
        """Same as setitem with the extra parameter substitute
        Args:
            substitute(bool): Substitute value with template if True
        """
        if substitute:
            value = self.substitute_string(value)

        self.__setitem__(key, value)

    def substitute_string(self, string: str):
        return Template(string).substitute(self.data)


class ReadonlyDict(collections.UserDict):
    def __setitem__(self, key, value):
        raise RuntimeError("Cannot modify read only dict")


class FrameworkDatabases(collections.UserDict):
    def __setitem__(self, key, value):
        raise RuntimeError("Cannot modify Databases")


class Framework:
    def __init__(
        self,
        name: str = None,
        *,
        homepath: str = None,
        home_structure=None,
        variables: Dict[str, Any] = None,
        version: str = None,
        config: Dict = None,
    ):

        # TODO: add param substitute_vars, to substitute variables on init?
        if variables is None:
            variables = {}

        if config is None:
            config = {}

        self.variables = FrameworkVariables(variables)

        self.name: str = name
        self.homepath: str = homepath
        self.config = config
        if version is None:
            version = "0.0.0"
        self.version: str = version
        # The framework can be initialized if it has a name or a home path
        self._can_be_initialized = name or homepath
        if self._can_be_initialized and self.homepath is None:
            self.homepath = os.path.expanduser(
                os.path.join("~", f".{HOMEPATH_NAME}", self.name)
            )

        if self.homepath:
            self.config_path = os.path.join(self.homepath, "config.toml")
        if home_structure is None:
            home_structure = []
        else:
            home_structure = normalize_home_structure(home_structure)
        self._home_structure = home_structure
        # self._plugin_managers: Dict[str, PluginManager] = {}
        self.plugin_managers = KoalakContainer()
        self._hook_managers: Dict[str, HookManager] = {}
        self._initialized: bool = False

        # --- variables ---
        # TODO: test home variable with unnamed framework
        if self.homepath:
            self.variables.data["home"] = self.homepath
        self.variables.data["userhome"] = os.path.expanduser("~")
        self.printer = Printer()

        # --- databases ---
        self.list_db: typing.Mapping[str : databases.ListDB] = FrameworkDatabases()
        self.dict_db: typing.Mapping[str : databases.DictDB] = FrameworkDatabases()

    def __repr__(self):
        if self.name:
            return f"<Framework [{self.name}]>"
        else:
            return f"<Framework>"

    def __str__(self):
        return self.__repr__()

    def fixture(self, function):
        pass

    def init(self):
        # We can init a framework if he has a name or a homepath
        if self._initialized:
            raise TypeError(f"Framework {self.name} already initialized")
        self._initialized = True

        # Check if it can be initialized
        if not self._can_be_initialized:
            raise TypeError(
                f"Framework need a name or a homepath in order to be initialized"
            )

        # Check if it's already initialized (in previous run)
        if not self._is_initialized():
            self._init_home()

        # Init plugins
        homepath_initialized = set()  # Don't initialize twice the same homepath!
        for plugin_manager in self.plugin_managers:
            homepath = plugin_manager.homepath
            plugin_manager.init(_homepath_initialized=homepath_initialized)
            homepath_initialized.add(homepath)

        # TODO: at the end init the framework database (version)

        if os.path.exists(self.config_path):
            self.config = toml.load(self.config_path)
        else:
            with open(self.config_path, "w") as f:
                toml.dump(self.config, f)

    def _is_initialized(self):
        return os.path.isdir(self.homepath)

    def _init_home(self, parent_path=None, nodes=None):
        """Recursive function"""
        # FIXME: add support for nodes as list of string and dict
        # FIXME: see _normalize_home_node and skipped tests
        # TODO: add support for tree ``tree -Jd``
        # TODO: add this as a standalone function
        if parent_path is None:
            os.makedirs(self.homepath)
            parent_path = self.homepath

        if nodes is None:
            nodes = self._home_structure

        for node in nodes:
            path = os.path.join(parent_path, node.name)
            if isinstance(node, FileDescription):
                # run before_normalize hooks actions
                for action in node.actions:
                    action.before_normalize(node, self)

                norimalized_file = home_structure.normalize_file_description(
                    node, parent_path, self
                )

                # run after_normalize hooks actions
                for action in node.actions:
                    action.after_normalize(norimalized_file)

                with open(norimalized_file.path, "w") as f:
                    f.write(norimalized_file.content)

                # run after_normalize hooks actions
                for action in node.actions:
                    action.after_creation(norimalized_file)

            elif isinstance(node, DirectoryDescription):
                os.mkdir(path)
                self._init_home(path, node.nodes)

    # @defaultargs(firstarg=str)
    def mkhkmanager(self, func, hook_name):
        pass

    def mkhookmanager(self, hook_name_or_func: Union[Callable, str]):
        def _mkhookmanager(spec_function):
            if hook_name in self._hook_managers:
                raise exceptions.HookManagerAlreadyExistException(
                    f"hook {hook_name} already exist in {self.name} framework"
                )

            hm = HookManager(hook_name, spec_function)
            self._hook_managers[hook_name] = hm
            return hm

        if isinstance(hook_name_or_func, str):
            hook_name = hook_name_or_func
        elif inspect.isfunction(hook_name_or_func):
            hook_name = hook_name_or_func.__name__
            return _mkhookmanager(hook_name_or_func)
        else:
            pass
            # TODO: raise not handled type

        return _mkhookmanager

    def mkpluginmanager(self, pluginmanager_name: str = None, homepath=None):

        if homepath is None and self.homepath:
            homepath = os.path.join(self.homepath, "plugins")
        elif homepath and self.homepath:
            homepath = os.path.join(self.homepath, homepath)

        pluginmanager = PluginManager(pluginmanager_name, homepath=homepath)
        self.plugin_managers._add_item(pluginmanager, pluginmanager_name)
        # self._plugin_managers[pluginmanager_name] = pluginmanager
        return pluginmanager

    def mkpluginmanager2(self, plugin_name: str):
        def _mkplugin(base_cls):
            if plugin_name in self._plugin_managers:
                raise exceptions.PluginManagerAlreadyExistException(
                    f"plugin manager {plugin_name} already exist"
                )
            pm = PluginManager(plugin_name, base_cls)
            self._plugin_managers[plugin_name] = pm

            base_cls._fwm = {"plugin_manager": pm}
            base_cls.__init_subclass__ = classmethod(__init_subclass__)
            return base_cls

        return _mkplugin

    def get_plugins(self, plugin_manager_name: str):
        from warnings import warn

        warn("get_plugins is deprecated use plugin_managers instead")

        return self.plugin_managers[plugin_manager_name].get_plugins()

    def get_plugin_manager(self, plugin_manager_name: str) -> PluginManager:
        from warnings import warn

        warn("get_plugin_manager is deprecated use plugin_managers[<name>] instead")
        return self.plugin_managers[plugin_manager_name]

    def get_hook_manager(self, hook_manager_name: str) -> HookManager:
        return self._hook_managers[hook_manager_name]

    def get_hook_managers(self):
        return list(self._hook_managers.values())

    def get_hook(self, hook_manager_name, hook_name):
        return self._hook_managers[hook_manager_name].get_hook(hook_name)

    def get_hooks(self, hook_manager_name):
        return self._hook_managers[hook_manager_name].get_hooks()

    def get_plugin(self, plugin_manager_name: str, plugin_name: str):
        # TODO: keep this function or not?
        return self.plugin_managers[plugin_manager_name].get_plugin(plugin_name)

    def create_list_db(
        self, name, *, path=None, type=None, unique=None
    ) -> databases.ListDB:
        # FIXME: init before or create database? or not important?
        # FIXME: check that db name is correct (no new lines..) as identifier
        if type is None:
            type = "json"
        elif type not in ["txt", "json"]:
            raise ValueError(f"Unknown type of db {type!r}")

        if path is None:
            path = f"$home/databases/list/{name}.{type}"

        path = self.substitute_string(path)
        dirname = os.path.dirname(path)
        os.makedirs(dirname, exist_ok=True)  # create DB directory

        if type == "json":
            cls_db = databases.JsonListDB
        elif type == "txt":
            cls_db = databases.TxtListDB
        else:
            # ignore this line in coverage because we check type in the begening of this function
            #  this line is here juste to help us avoid programing errors
            raise ValueError(f"Unknown type of db {type!r}")  # pragma: no cover
        db = cls_db(path, unique=unique)

        self.list_db.data[name] = db
        return db

    def create_dict_db(
        self, name, *, path=None, type=None, unique=None, **kwargs
    ) -> databases.DictDB:
        """
        Args:
            kwargs: kwargs argument to pass to DictDB
        """
        # FIXME: init before or create database? or not important?
        # FIXME: check that db name is correct (no new lines..) as identifier
        if type is None:
            type = "json"
        elif type not in ["txt", "json"]:
            raise ValueError(f"Unknown type of db {type!r}")

        if path is None:
            path = f"$home/databases/dict/{name}.{type}"

        path = self.substitute_string(path)
        dirname = os.path.dirname(path)
        os.makedirs(dirname, exist_ok=True)  # create DB directory

        if type == "json":
            cls_db = databases.JsonDictDB
        elif type == "txt":
            cls_db = databases.TxtDictDB
        else:
            # ignore this line in coverage because we check type in the begening of this function
            #  this line is here juste to help us avoid programing errors
            raise ValueError(f"Unknown type of db {type!r}")  # pragma: no cover
        db = cls_db(path, unique=unique, **kwargs)

        self.dict_db.data[name] = db
        return db

    def print_info(self, details=True):
        """Function to debug and see main information about this framewoork"""
        # TODO: coloring add coloring.print_title
        coloring.print_green(f"Framework {self.name}")

        print("Plugin managers:")
        for pm in self.plugin_managers:
            print(f"\t{pm.name} ({len(pm)})")

        print("Hooks:")
        for hook_name, hook in self._hook_managers.items():
            print(f"\t{hook_name} ({len(hook)})")
        print("Variables:")
        for varname, value in self.variables.items():
            print(f"\t{varname}: {value}")

        # databases
        print("Databases:")
        print("\tList DB:")
        for db_name, db in self.list_db.items():
            print(f"\t\t{db_name} ({len(db)}) [{db.uri}]")

        print("\tDict DB:")
        for db_name, db in self.dict_db.items():
            print(f"\t\t{db_name} ({len(db)}) [{db.uri}]")
        # TODO: plugin_managers/hooks
        # TODO: change naming

    def uninstall(self):
        """Undo every thing init did"""
        # TODO: can't be called without init being called
        # TODO: test! test plugins (they might have their own path)
        # Remove the home path
        shutil.rmtree(self.homepath)

        # undo file actions
        # FIXME: must be recursive! implement iternodes
        for node in self._home_structure:
            if isinstance(node, FileDescription):
                for action in node.actions:
                    action.uninstall(node, self)

    # ========= #
    # VARIABLES #
    # ========= #
    def substitute_string(self, string: str):
        return self.variables.substitute_string(string)


class FrameworkManager(KoalakContainer):
    """This class represent in a way the library Koalak"""

    def __init__(self):
        super().__init__()
        self._unique_framework_name_counter = 0

    def mkframework(
        self,
        framework_name: str = None,
        *,
        homepath=None,
        home_structure=None,
        variables=None,
        version: str = None,
        config: Dict = None,
    ) -> Framework:
        if framework_name and framework_name in self:
            raise exceptions.FrameworkAlreadyExistException(
                f"Framework {framework_name} already exist"
            )

        framework = Framework(
            framework_name,
            homepath=homepath,
            home_structure=home_structure,
            variables=variables,
            version=version,
            config=config,
        )

        self._add_item(framework, framework_name)
        return framework

    def get_framework(self, framework_name: str) -> Framework:
        import warnings

        warnings.warn(
            "get_framework deprecated use frameworks[<name>] instead",
            DeprecationWarning,
        )
        return self[framework_name]

    def get_frameworks(self) -> List[Framework]:
        import warnings

        warnings.warn(
            "get_frameworks deprecated use frameworks instead", DeprecationWarning
        )
        return list(self)

    def generate_unique_framework_name(self):
        while True:
            self._unique_framework_name_counter += 1
            counter = self._unique_framework_name_counter
            name = f"unique_framework_name_{counter}"
            if name not in self:
                return name

    def get_unique_framework_name(self) -> str:
        import warnings

        warnings.warn(
            "get_unique_framework_name deprecated use generate_unique_framework_name instead",
            DeprecationWarning,
        )
        return self.generate_unique_framework_name()


frameworks = FrameworkManager()



import functools

# TODO: create ObjectStore to store attributes in object with a specific namespace
import inspect
import logging

logger = logging.getLogger("pff")
logger.setLevel(logging.DEBUG)
logger.setLevel(logging.CRITICAL)


class ObjectStore:
    def __init__(self):
        pass

    def has(self, attribute):
        pass

    def get(self):
        pass

    def set(self):
        pass


def factory_dispatcher__init__(cls):
    # Create the new __init__
    # we will call functools.wraps after
    def dispatcher__init__(self, *args, **kwargs):
        for __init__ in cls.__inits__:
            __init__(self, *args, **kwargs)

        for __post_init__ in cls.__post_inits__:
            __post_init__(self)

    # Wraps functions for signature
    # if the cls use the object.__init__ we have
    # to replace it signature with a signature without *args, **kwargs
    if cls.__init__ == object.__init__:

        def dummy__init__(self):
            pass

        wrapped_signature = dummy__init__
    else:
        wrapped_signature = cls.__init__
    dispatcher__init__ = functools.wraps(wrapped_signature)(dispatcher__init__)
    return dispatcher__init__


def add_post_init(post_init):
    """A decorator for class to add a function to be called
    after init.
    The decorator will create an __init__ if doesn't exist
    if an __init__ exist it will replace it with a an other one
    called __dispatcher_init__ that will call the old __init__
    - possible to have many post_inits

    Example:
        def addx(self):
            self.x = True

        @add_post_init(addx)
        class A:
        def __init__():
            self.a = True

        assert a.a is True
        assert a.x is True

    Args:
        post_init: function with one positional argument (self)

    # TODO: check that when creating an init (cls don't have one)
    # we create a init with a simple signature not *args, **kwargs
    names: in the cls class (TODO: change it in a namespace for this package)
        - __inits__
        - __post_inits__
    """
    # TODO: support multiple functions at once *init_functions
    def decorator(cls):
        if not hasattr(cls, "__inits__"):
            # create a new __init__
            dispatcher__init__ = factory_dispatcher__init__(cls)
            # replace the original __init__ by the dispathcer__init__
            cls.__inits__ = [cls.__init__]
            cls.__post_inits__ = []
            cls.__init__ = dispatcher__init__
        # add the post_init function
        # this function will be called by the dispathcer__init__
        cls.__post_inits__.append(post_init)

        return cls

    return decorator


# TODO: add order
def addinit(added__init__, *added__inits__, runfirst=False):
    # TODO: test this class
    # TOOD: use the factory dispatcher__init__
    # TODO: test addinit with add_postinit
    added__inits__ = [added__init__, *added__inits__]

    def dispatcher__init__(self, *args, **kwargs):
        for __init__ in dispatcher__init__.__inits__:
            __init__(self, *args, **kwargs)

    def decorator(cls):
        try:
            # Check if addinit was already called or not
            # if addinit is called, the __init__ will
            # have the attributes __inits__
            # Warning: check the __init__ of the class (__dict__['__init__'])
            # not the __init__ cls.__init__ (it could be the parent __init__)
            __inits__ = cls.__dict__["__init__"].__inits__
        except (KeyError, AttributeError):
            # if it's the first time we call addinit
            # we dynamically add __inits__ in cls.__init__
            __inits__ = [cls.__init__]
            dispatcher__init__.__inits__ = __inits__
            cls.__init__ = dispatcher__init__

        # Handle the order of execution
        if runfirst:
            __inits__ = added__inits__ + __inits__
        else:
            __inits__ += added__inits__
        return cls

    return decorator


# TODO: add way to specify runfirst
def persistent_addinit(added_init, runfirst=False):
    def added_init_decorator(cls):
        return addinit(added_init)(cls)

    return added_init_decorator


def defaultdecorator(_addattr=None, *, decorated=None, firstarg=None):
    """Decorator for decorators to allow them to be called with or without
    arguments.

    Example:
        @defaultdecorator
        def decorator(obj, msg="Hello"):
            print(msg)
            return obj

        # Now decorator can be called with (@decorator(msg="somthing")) or
        # without arguments (@decorator)

        @decorator
        class A:
            pass

        @decorator(msg="Good Bye")
        class A:
            pass

    This decorator itself can be called with or without arguments.

    By default only keyword arguments are allowed for the decorated decorator
    to enable positional arguments you have to specify either the type of
    the first argument or the type of the decorated object (by the decorated
    decorator) to resolve ambiguity.



    Args:
        decorated: type of the expected decorated object (by the decorated decorator)
            to resolve ambiguity
        firstarg: type of the first argument of the decorated decorator to resolve
            ambiguity

    """
    if _addattr is not None:
        # defaultdecorator called without kwargs. Example
        #  @defaultdecorator
        #  def f():
        #      pass
        @functools.wraps(_addattr)
        def new_addattr(obj=None, **kwargs):
            # no *args, the decorator must be called with kwargs only
            if obj is None:
                # obj is None => @decorator(a=2, x=3)
                print("obj is None => @decorator(a=2, x=3)")

                def deco1(cls):
                    return _addattr(cls, **kwargs)

                return deco1
            else:
                # obj is not None => @decorator without args
                print("obj is not None => @decorator")
                return _addattr(obj, **kwargs)

        return new_addattr
    else:
        # defaultdecorator called with kwargs to enable positional arguments

        if decorated is not None and firstarg is not None:
            raise TypeError(f"You can't use both 'decorated' and 'firstarg' keyword")

        if decorated is not None:
            # Normalize decorated to a tuple
            if not isinstance(decorated, tuple):
                decorated = tuple([decorated])

            def isarg_for_decorator(decorated_obj):
                return not isinstance(decorated_obj, decorated)

            print("arg decorated", decorated)

        elif firstarg is not None:
            # Normalize decorated to a tuple
            if not isinstance(firstarg, tuple):
                firstarg = tuple([firstarg])

            def isarg_for_decorator(decorated_obj):
                """Tell if the argument is for the decorator or if
                it's the object to decorate"""
                return isinstance(decorated_obj, firstarg)

            print("arg firstarg", firstarg)

        else:
            pass
            print("NOTHING GIVEN PROBLEM")
            raise TypeError()
            # TODO: raise error?

        def _defaultdecorator(_addattr):
            @functools.wraps(_addattr)
            def new_addattr(obj_or_arg=None, *args, **kwargs):
                if isarg_for_decorator(obj_or_arg):
                    # obj is None => @decorator(a=2, x=3)
                    print("obj is None => @decorator(a=2, x=3)")
                    if obj_or_arg:
                        args = [obj_or_arg] + list(args)
                    print("args", args)
                    print("kwargs", kwargs)

                    def deco1(cls):
                        return _addattr(cls, *args, **kwargs)

                    return deco1
                else:
                    # obj is not None => @decorator without args
                    print("obj is not None => @decorator")
                    print("args", args)
                    print("kwargs", kwargs)
                    return _addattr(obj_or_arg, *args, **kwargs)

            return new_addattr

        return _defaultdecorator


def addattr(obj=None, *, attr="x", value=5):
    def payload(obj):
        setattr(obj, attr, value)
        return obj

    if obj is None:
        # called with args @addattr() @addattr(x=2)
        def decorator(obj):
            return payload(obj)

        return decorator
    else:
        # called without args @addattr
        return payload(obj)


def is_future_method(function):
    """Tell if this function will be a method based on if it's first argument is named 'self'"""
    # this function is used in "optionalargs"
    try:
        return inspect.getfullargspec(function).args[0] == "self"
    except IndexError:
        return False


def optionalargs(decorator_to_decorate, *, firstarg=None):
    # TODO: decorator as a class
    # TODO: yassmine: add option to disable "ismethod" guessing, and to tell that it's a method
    """Decorator for decorators to allow them to be called with or without
    arguments.

    Example:
        ```
        @optionalargs
        def decorator(obj, msg="Hello"):
            print(msg)
            return obj

        # Now the decorator can be called with or without arguments

        # without arguments:
        @decorator
        def f():
            pass

        # with arguments
        @decorator(msg="Good Bye")
        def f():
            pass
        ```
    optionalargs can be called with or without arguments.

    By default only keyword arguments are allowed for the decorated decorator
    to enable positional arguments you have to specify either the type of
    the first argument or the type of the decorated object (by the decorated
    decorator) to resolve ambiguity.



    Args:
        decorated: type of the expected decorated object (by the decorated decorator)
            to resolve ambiguity
        firstarg: type of the first argument of the decorated decorator to resolve
            ambiguity

    """
    logger.error(
        "optionalargs called with decorator=%s firstarg=%r",
        decorator_to_decorate,
        firstarg,
    )

    obj_is_method = is_future_method(decorator_to_decorate)

    def _refactored(obj_or_arg, args, kwargs, self=None):
        pre_args = []
        if self is not None:  # this was if self and I took 3 hours to debug it!
            pre_args.append(self)

        # There are 3 possibility for the first argument
        #  None => decorator called with brackets =>
        if obj_or_arg is None or (
            firstarg is not None and isinstance(obj_or_arg, firstarg)
        ):
            logger.error(
                "decorator %s called with brackets", decorator_to_decorate.__name__
            )
            logger.error("obj_or_arg %r args %s kwargs %s", obj_or_arg, args, kwargs)
            # decorator called with brackets
            #  @decorateme()
            #  def f():
            #      pass
            # or:
            #  @decorateme(arg=value)
            #  def f():
            #      pass
            arg = obj_or_arg
            del obj_or_arg

            if arg is not None:
                args = (arg, *args)

            def decorator(obj):
                return decorator_to_decorate(*pre_args, obj, *args, **kwargs)

            return decorator
        else:
            # decorator called without braketcs. Example
            #  @decorateme
            #  def f():
            #      pass
            obj = obj_or_arg
            del obj_or_arg

            logger.error(
                "decorator %s called without brackets", decorator_to_decorate.__name__
            )
            return decorator_to_decorate(*pre_args, obj)

    if obj_is_method:
        # if it's a method we must add self in the signature
        @functools.wraps(decorator_to_decorate)
        def decorated_decorate_me(self, obj_or_arg=None, *args, **kwargs):
            return _refactored(obj_or_arg, args, kwargs, self=self)

    else:

        @functools.wraps(decorator_to_decorate)
        def decorated_decorate_me(obj_or_arg=None, *args, **kwargs):
            return _refactored(obj_or_arg, args, kwargs)

    return decorated_decorate_me


# decorate optionalargs with itself
optionalargs = optionalargs(optionalargs)
logger.error("after init optionalargs")


"""
@optionalargs
def decorateme_addattr(obj, *, attr="x", value=5):
    setattr(obj, attr, value)
    return obj

@decorateme_addattr
class A:
    pass


assert A.x == 5


@addattr(attr="y", value=2)
class A:
    pass


assert A.y == 2




def addattr(*, attr="x", value=5):
    def decorator(obj):
        setattr(obj, attr, value)
        return obj
    return decorator
"""



import argparse

import koalak

# Creating framework and pluginmanager
wordlistools = koalak.mkframework("wordlistools")
tools = wordlistools.mkpluginmanager("tools")


# Creating ou BasePlugin with constraints on the attribute
@tools.mkbaseplugin
class BaseTool:
    type = tools.attr(type=str, choices=["filters", "modifiers"])
    func = tools.attr()


# Adding Modifiers Plugin tools
class LowerTool(BaseTool):
    type = "modifiers"
    name = "lower"

    def func(self, string):
        return string.lower()


class UpperTool(BaseTool):
    type = "modifiers"
    name = "upper"

    def func(self, string):
        return string.upper()


class InvertTool(BaseTool):
    type = "modifiers"
    name = "invert"

    def func(self, string):
        return string[::-1]


# Adding Filters Plugin tools
class IsdigitTool(BaseTool):
    type = "filters"
    name = "isdigit"

    def func(self, string):
        return string.isdigit()


class IsupperTool(BaseTool):
    type = "filters"
    name = "isupper"

    def func(self, string):
        return string.isupper()


# Main program logic
def run_wordlistools(function_name, filepath):
    # Get the plugin class with its name
    Tool = tools[function_name]
    tool = Tool()

    if tool.type == "filters":
        with open(filepath) as f:
            for line in f:
                line = line[:-1]
                if tool.func(line):
                    print(line)
    elif tool.type == "modifiers":
        with open(filepath) as f:
            for line in f:
                line = line[:-1]
                print(tool.func(line))


if __name__ == "__main__":
    # Get the name of all our tools by iterating the PluginManager
    tool_names = [tool.tool_name for tool in tools]

    parser = argparse.ArgumentParser()
    parser.add_argument("tool", help="Tool to use", choices=tool_names)
    parser.add_argument("filepath", help="Wordlist path")
    args = parser.parse_args()

    run_wordlistools(args.tool_name, args.filepath)



import koalak
import pytest
from koalak import exceptions


def nothing(**args):
    pass


# TODO: test CRUD on plugins/frameworks
@pytest.mark.skip
def test_plugin():
    framework = koalak.mkframework()

    @framework.mkpluginmanager("test")
    class BaseTest:
        pass

    assert framework.get_plugins("test") == []
    pm_test = framework.get_plugin_manager("test")
    assert pm_test.get_plugins() == []


@pytest.mark.skip
def test_sub_plugin():
    framework = koalak.mkframework()

    @framework.mkpluginmanager("collector")
    class BaseCollector:
        pass

    class SimpleCollector(BaseCollector):
        name = "simple"

    assert framework.get_plugin("collector", "simple") is SimpleCollector
    pm_collector = framework.get_plugin_manager("collector")
    assert pm_collector.get_plugin("simple") is SimpleCollector

    assert pm_collector.get_plugins() == [SimpleCollector]


@pytest.mark.skip
def test_two_plugin_managers():
    framework = koalak.mkframework()

    @framework.mkpluginmanager("collector")
    class BaseCollector:
        pass

    class XCollector(BaseCollector):
        name = "x"

    class YCollector(BaseCollector):
        name = "y"

    @framework.mkpluginmanager("installer")
    class BaseInstaller:
        pass

    class XInstaller(BaseInstaller):
        name = "x"

    class YInstaller(BaseInstaller):
        name = "y"

    assert framework.get_plugins("collector") == [XCollector, YCollector]
    assert framework.get_plugins("installer") == [XInstaller, YInstaller]

    assert framework.get_plugin("collector", "x") is XCollector
    assert framework.get_plugin("collector", "y") is YCollector
    assert framework.get_plugin("installer", "x") is XInstaller
    assert framework.get_plugin("installer", "y") is YInstaller

    pm_collector = framework.get_plugin_manager("collector")
    assert pm_collector.get_plugin("x") is XCollector
    assert pm_collector.get_plugin("y") is YCollector

    pm_installer = framework.get_plugin_manager("installer")
    assert pm_installer.get_plugin("x") is XInstaller
    assert pm_installer.get_plugin("y") is YInstaller


@pytest.mark.skip
def test_error_plugin_manager_already_exist():
    framework = koalak.mkframework()

    @framework.mkpluginmanager("a")
    class A:
        pass

    with pytest.raises(exceptions.PluginManagerAlreadyExistException):

        @framework.mkpluginmanager("a")
        class B:
            pass


@pytest.mark.skip
def test_error_plugin_alread_exist():
    framework = koalak.mkframework()

    @framework.mkpluginmanager("plugin")
    class BasePlugin:
        pass

    class APlugin(BasePlugin):
        name = "a"

    with pytest.raises(exceptions.PluginAlreadyExistException):

        class AagainPlugin(BasePlugin):
            name = "a"


@pytest.mark.skip
def test_get_frameworks():
    framework1 = koalak.mkframework()
    framework2 = koalak.mkframework()

    frameworks = koalak.get_frameworks()
    assert framework1 in frameworks
    assert framework2 in frameworks



"""These classes are here to create the home_structure when initiating a new framework

"""
import os
from string import Template
from typing import Any, Dict, List, Union

import attr

"""
Functions to create directory (and files) structure easily

see also:
- https://stackoverflow.com/questions/26288551/render-json-to-directory-structure-in-python

"""


def _file_description_action_converter(actions):
    if isinstance(actions, str):
        return [file_actions_map[actions]()]
    elif isinstance(actions, FileAction):
        return [actions]
    elif isinstance(actions, list):
        new_actions = []
        for action in actions:
            if isinstance(action, str):
                action = FileAction()
            elif isinstance(action, FileAction):
                pass
            else:
                raise TypeError(
                    f"actions must be of type 'str' or {FileAction.__name__!r}"
                )
            new_actions.append(action)
        return actions
    else:
        raise TypeError(
            f"actions must be of type 'str' or {FileAction.__name__!r} or list"
        )


@attr.s
class FileDescription:
    """Used to describe file to create the home structure"""

    name = attr.ib()
    src = attr.ib(default=None)
    content = attr.ib(default="")
    substitute = attr.ib(default=False)
    actions = attr.ib(
        default=attr.Factory(list), converter=_file_description_action_converter
    )


@attr.s
class NormalizedFile:
    """Normalize FileDescription"""

    path = attr.ib()
    content = attr.ib()
    framework = attr.ib()


def normalize_file_description(
    file_description: FileDescription, parent_path, framework
):
    # FIXME: handle the case of big files (copy by chunk)? and test it
    # FIXME: when we have a source file? do we need to copy the permissions?
    # get the content of the file
    if file_description.src:
        # from a source file
        with open(file_description.src) as f:
            content = f.read()
    else:
        content = file_description.content
    # substitute variables
    # TODO: add general substitute variable in the framework to default to
    if file_description.substitute:
        content = framework.substitute_string(content)

    path = os.path.join(parent_path, file_description.name)
    return NormalizedFile(path, content, framework)


@attr.s
class DirectoryDescription:
    """Used to describe directory to create the home structure"""

    name = attr.ib()
    nodes = attr.ib(default=attr.Factory(list))


# ======= #
# ACTIONS #
# ======= #


class FileAction:
    """Hook when creating files"""

    def before_normalize(self, file_description: FileDescription, framework):
        pass

    def after_normalize(self, normalized_file: NormalizedFile):
        pass

    def after_creation(self, normalized_file: NormalizedFile):
        pass

    def uninstall(self, file: FileDescription, framework):
        pass


class BashrcFileAction(FileAction):
    """This file will be executed on the bashrc"""

    _add_to_bashrc_template = """\n# Execute the '{basename}' bash file of '{framework_name}'
if [ -f {path} ]; then
    . {path}
fi
"""

    def __init__(self, bashrcpath=None):
        if bashrcpath is None:
            bashrcpath = os.path.join(os.path.expanduser("~"), ".bashrc")
        self.bashrcpath = bashrcpath
        self.added_bashrc = None

    def _get_add_to_bashrc(self, file: NormalizedFile):
        basename = os.path.basename(file.path)
        return self._add_to_bashrc_template.format(
            basename=basename, path=file.path, framework_name=file.framework.name
        )

    def after_normalize(self, file: NormalizedFile):

        with open(self.bashrcpath) as f:
            bashrc = f.read()

        add_to_bashrc = self._get_add_to_bashrc(file)

        if add_to_bashrc not in bashrc:
            with open(self.bashrcpath, "a") as f:
                f.write(add_to_bashrc)
        # store added_bashrc to be able to uninstall it
        self.added_bashrc = add_to_bashrc

    def uninstall(self, file_description, framework):
        """Remove the added lines in the bashrc file"""
        # FIXME: uninstall can be called in the same execution when the file is created
        # or it can be called in an other execution, so the self.added_bashrc could not be
        # intiialised and be None, added_bashrc must be deduced.
        with open(self.bashrcpath) as f:
            bashrc = f.read()
        # if self.added_bashrc in bashrc:
        #    bashrc = bashrc.replace(self.added_bashrc, "")

        with open(self.bashrcpath, "w") as f:
            f.write(bashrc)


file_actions_map = {"bashrc": BashrcFileAction}
"""
Differents actions:
- bashrc: point ~/.bashrc to execute this script
- for eachfile
    - run prehook on File
    - normalize File to NormalizedFile
    - run hook on NormalizedFile

init_home steps
- normalize home to a list of File and Directory (recursively)
-


"""


# TODO: add uninstall (remove every thing init did) (folders, databases, actions, ...)
def normalize_home_structure(
    nodes,
) -> List[Union[FileDescription, DirectoryDescription]]:
    if isinstance(nodes, list):
        return _normalize_home_nodes_list(nodes)
    elif isinstance(nodes, dict):
        return _normalize_home_nodes_dict(nodes)
    return nodes


def _normalize_home_nodes_list(
    nodes: List[str],
) -> List[Union[FileDescription, DirectoryDescription]]:
    normalized_nodes = []
    for node in nodes:
        if isinstance(node, str):
            # TODO: check if it contains "/"
            if node[-1] == "/":
                new_node = DirectoryDescription(node[:-1])
            else:
                new_node = FileDescription(node)
        else:
            new_node = node
        normalized_nodes.append(new_node)
    return normalized_nodes


def _normalize_home_nodes_dict(
    nodes: Dict[str, Any]
) -> List[Union[FileDescription, DirectoryDescription]]:
    normalized_nodes = []
    for name, value in nodes.items():
        if isinstance(value, str):
            new_node = FileDescription(name, content=value)
        elif isinstance(value, dict):
            new_node = DirectoryDescription(name, _normalize_home_nodes_dict(value))
        if isinstance(node, str):
            # TODO: check if it contains "/"
            if node[-1] == "/":
                new_node = DirectoryDescription(node[:-1])
            else:
                new_node = FileDescription(node)
        else:
            new_node = node
        normalized_nodes.append(new_node)
    return normalized_nodes


"""
class File:
    def __init__(self, name: str, src=None, content=None):
        self.name = name
        self.src = src
        if content is None:
            content = ""
        self.content = content


class D:
    def __init__(self, name, *nodes):
        self.name = name
        self.nodes = nodes
"""



import os

import koalak
import pytest
from koalak.utils import temp_pathname


def test_simple():
    """Test framework can be created"""
    koalak.mkframework()


def test_double_init():
    """If we call init twice raise Error"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        framework.init()
        with pytest.raises(TypeError):
            framework.init()


def test_homepath_is_created():
    """Homepath is created after init"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        assert not os.path.exists(pathname)
        framework.init()
        assert os.path.isdir(pathname)


def test__repr__and__str__():
    framework = koalak.mkframework()
    assert repr(framework) == str(framework) == "<Framework>"

    framework = koalak.mkframework("wordlistools")
    assert repr(framework) == str(framework) == "<Framework [wordlistools]>"


# TODO: add constraint on the name of the framework (no new line...)
# TODO: add description of the framework and plugins
# TODO: many pluginsmanagers without name?
# TODO: 2 pluginsmanagers sharing the same folder?
# TODO: When loading a new plugin from home can we sandbox it?
#   to load only this plugin and not an other? or malicious code?
# TODO: add a way te remove a framwork? (from the list to free the memory)


def test_homepath_plugins_alone():
    """Test that folder plugins is created"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        framework.mkpluginmanager()
        assert not os.path.exists(pathname)
        framework.init()
        assert os.path.isdir(pathname)
        assert os.path.isdir(os.path.join(pathname, "plugins"))


def test_homepath_plugins_with_homepath():
    """Test that we can chose the folder name of plugins"""
    with temp_pathname() as pathname:
        framework = koalak.mkframework(homepath=pathname)
        plugins = framework.mkpluginmanager(homepath="tools")
        assert not os.path.exists(pathname)
        framework.init()
        assert os.path.isdir(pathname)
        assert not os.path.isdir(os.path.join(pathname, "plugins"))
        assert os.path.isdir(os.path.join(pathname, "tools"))


def test_unique_framework_name():
    """Unique name are correctly generated"""
    for _ in range(10):
        name = koalak.generate_unique_framework_name()
        framework = koalak.mkframework(name)
        assert koalak.frameworks[name] is framework



"""THIS CODE IS A WORK IN PROGRESS"""
import collections
import datetime
import inspect
import logging
import typing

import attr
import coloring
from koalak.consts import PACKAGE_NAME
from koalak.decorators import optionalargs

logger = logging.getLogger("koala")
logger.setLevel(logging.DEBUG)

try:
    typping_ForwardRef = typing.ForwardRef
except AttributeError:
    typping_ForwardRef = typing._ForwardRef  # noqa


"""
RelationalDB is a wrapper for sqlalchemy and attr to have
a relational database with a better API and abstraction
- no ID
- specify relations with annoation (List["Animal"], Animal, ...)
- use attr to create __init__, __repr__, ...
See examples in test_relationaldb.py



"""


@attr.s
class BuildingRelationalAttribute:
    """Helper cls to build relational attribute"""

    name: str = attr.ib()
    many: bool = attr.ib(default=None)
    atomic_type_name: str = attr.ib(
        default=None
    )  # type of the referenced table (without the List)
    referenced_by: "BuildingRelationalAttribute" = attr.ib(default=None)
    attr_attribute = attr.ib(default=None)

    # this attribute is set to True by the class that contain this
    #  attribute, sins it is possible for an other class to create
    #  an attribute with ref (having exist=True we are sure that
    #  the attribute exist and it's not just a reference)
    exist = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.name == "age":
            coloring.print_bred("AGE IS BUILT")
            print(self)

    def update(
        self,
        *,
        many=None,
        atomic_type_name=None,
        referenced_by=None,
        attr_attribute=None,
        exist=None,
    ):
        for varname in [
            "exist",
            "many",
            "atomic_type_name",
            "referenced_by",
            "attr_attribute",
        ]:
            var = locals()[varname]
            if var is not None:
                attr = getattr(self, varname)
                if attr is not None:
                    # attr not None and var not None! they must be the same
                    if var is not attr:
                        raise ValueError(
                            f"Trying to update {varname} but it's already set to {attr} "
                        )
                else:
                    setattr(self, varname, var)


@attr.s
class BuildingClass:
    """Helper class to postpone building the ORM classes"""

    name: str = attr.ib()
    cls = attr.ib(default=None, init=False)
    table_name: str = attr.ib(default=None, init=False)
    attributes: typing.Dict[str, BuildingRelationalAttribute] = attr.ib(
        factory=dict, init=False
    )
    orm_cls = attr.ib(default=None, init=False)
    # this attribute is set to True by the class that contain this
    #  attribute, sins it is possible for an other class to create
    #  an attribute with ref (having exist=True we are sure that
    #  the attribute exist and it's not just a reference)
    exist = attr.ib(default=None)

    def update(self, *, cls=None, table_name=None, orm_cls=None, exist=None):
        for varname in ["cls", "table_name", "orm_cls", "exist"]:
            var = locals()[varname]
            if var is not None:
                attr = getattr(self, varname)
                if attr is not None:
                    # attr not None and var not None! they must be the same
                    if var is not attr:
                        raise ValueError(
                            f"Trying to update {varname} but it's already set to {attr} "
                        )
                else:
                    setattr(self, varname, var)

    def get_or_create_attribute(
        self, attribute_name: str
    ) -> BuildingRelationalAttribute:
        if attribute_name in self.attributes:
            return self.attributes[attribute_name]
        else:
            attribute = BuildingRelationalAttribute(attribute_name)
            self.attributes[attribute_name] = attribute
            return attribute


@attr.s
class AttributeMetadata:
    """Additional metadata to attr (use sqlalchemy attributes and RelatinalDB attributes)"""

    unique: bool = attr.ib(default=None, kw_only=True)


@attr.s
class AttrMetadata:
    """Metadata used in attr.ib(metadata=...) that hold extra information about DB"""

    name: str = attr.ib(default=None)
    unique: bool = attr.ib(default=None)
    many: bool = attr.ib(default=None)
    atomic_type_name: str = attr.ib(
        default=None
    )  # type of the referenced cls (without the List)
    atomic_type = attr.ib(default=None)
    ref: str = attr.ib(default=None)
    attr_attribute = attr.ib(default=None)
    exist = attr.ib(default=None)
    referenced_by: "AttrMetadata" = attr.ib(default=None)

    def __attrs_post_init__(self):
        if self.unique is None:
            self.unique = False

    def __getattr__(self, item):
        if item in ["type"]:
            return getattr(self.attr_attribute, item)
        return self.__getattribute__(item)

    def update(
        self,
        *,
        many: bool = None,
        atomic_type_name: str = None,
        atomic_type=None,
        referenced_by=None,
        attr_attribute=None,
        exist: bool = None,
        name: str = None,
        ref: str = None,
    ):
        for varname in [
            "exist",
            "many",
            "atomic_type_name",
            "atomic_type",
            "referenced_by",
            "attr_attribute",
            "name",
            "ref",
        ]:
            var = locals()[varname]
            if var is not None:
                attr = getattr(self, varname)
                if attr is not None:
                    # attr not None and var not None! they must be the same
                    if var is not attr:
                        raise ValueError(
                            f"Trying to update {varname} but it's already set to {attr} "
                        )
                else:
                    setattr(self, varname, var)


@attr.s
class ClsMetadata:
    """Helper class to postpone building the ORM classes"""

    name: str = attr.ib()
    cls = attr.ib(default=None, init=False)
    table_name: str = attr.ib(default=None, init=False)
    attributes: typing.Dict[str, AttrMetadata] = attr.ib(factory=dict, init=False)
    # this attribute is set to True by the class that contain this
    #  attribute, sins it is possible for an other class to create
    #  an attribute with ref (having exist=True we are sure that
    #  the attribute exist and it's not just a reference)
    exist = attr.ib(default=None)

    def __len__(self):
        return len(self.attributes)

    def __contains__(self, item):
        return self.attributes.__contains__(item)

    def __getitem__(self, item) -> AttrMetadata:
        return self.attributes.__getitem__(item)

    def update(self, *, cls=None, table_name=None, exist=None):
        for varname in ["cls", "table_name", "exist"]:
            var = locals()[varname]
            if var is not None:
                attr = getattr(self, varname)
                if attr is not None:
                    # attr not None and var not None! they must be the same
                    if var is not attr:
                        raise ValueError(
                            f"Trying to update {varname} but it's already set to {attr} "
                        )
                else:
                    setattr(self, varname, var)

    def get_or_create_attribute(self, attribute_name: str) -> AttrMetadata:
        if attribute_name in self.attributes:
            return self.attributes[attribute_name]
        else:
            attribute = AttrMetadata()
            attribute.update(name=attribute_name)
            self.attributes[attribute_name] = attribute
            return attribute


class DBMetadata:
    def __init__(self):
        self.classes = {}

    def register(self, cls):
        self.decorate_with_attr(cls)
        cls_metadata = self.get_or_create_cls_metadata(cls.__name__)
        cls_metadata.update(exist=True, cls=cls)

        attr_attributes = cls.__attrs_attrs__
        for attr_attribute in attr_attributes:
            self.register_attribute(cls_metadata, attr_attribute)

    def register_attribute(self, cls_metadata: ClsMetadata, attr_attribute):
        coloring.print_bgreen("Register attribute", attr_attribute.name)
        attribute_name = attr_attribute.name
        # Create the attribute if not exist
        attribute = cls_metadata.get_or_create_attribute(attribute_name)
        attribute.update(attr_attribute=attr_attribute)

        atomic_type = self.get_atomic_type(attr_attribute.type)
        if isinstance(atomic_type, str):
            atomic_type_name = atomic_type
        else:
            atomic_type_name = atomic_type.__name__

        metadata = attr_attribute.metadata[PACKAGE_NAME]
        attribute.update(
            exist=True,
            atomic_type_name=atomic_type_name,
            atomic_type=atomic_type,
            ref=metadata.ref,
        )

        # If the type of this attribute is an other class (string referencing it)
        #  register the class
        if isinstance(atomic_type, str):
            self.get_or_create_cls_metadata(atomic_type_name)

        # Handle ref - the attribute reference an other attribute
        if attribute.ref:
            # Check that type is not built-in
            #  it's not possible to reference builtin types like str
            if self.is_builtin(attribute.atomic_type):
                # TODO: test
                raise ValueError(
                    f"ref cannot reference built-in type {attribute.atomic_type!r}"
                )

            # Create the attribute if not exit
            ref_cls_metadata = self.get_or_create_cls_metadata(atomic_type_name)
            ref_attribute = ref_cls_metadata.get_or_create_attribute(attribute.ref)

            # Check if the attribute already exist
            if ref_cls_metadata.exist and not ref_attribute.exist:
                # if the class exist (already registered) the attribute must exist
                raise ValueError(
                    f"{cls_metadata.name}.{attribute.name} reference non existing attribute {attribute.name} in {attribute.atomic_type_name}"
                )

            # update attribute
            ref_attribute.update(
                referenced_by=attribute,
                atomic_type=cls_metadata.cls,
                atomic_type_name=cls_metadata.name,
            )

    def get_atomic_type(self, type):
        return type

    def handle_type(self, attribute: AttrMetadata, cls_metadata: ClsMetadata):
        """Get the atomic_type and 'many'"""
        type = attribute.type
        many = False
        if isinstance(type, str):
            atomic_type_name = type
            atomic_type = type
        elif isinstance(type, typing.GenericMeta):
            if type.__name__ == "List":
                many = True
                # check if List is empty (ex: List instead of List[Animal])
                if not attr_type.__args__:  # noqa
                    raise ValueError(
                        f"List annotation for {cls_metadata.name}.{attribute.name} can not be empty"
                    )
                attr_type = attr_type.__args__[0]  # noqa

                # Check if the argument of List is a string
                if isinstance(attr_type, typping_ForwardRef):
                    attr_type = attr_type.__forward_arg__
            else:
                raise ValueError(
                    f"Support only List annotation, don't support {attr_type}"
                )
        else:
            atomic_type_name = atomic_type.__name__

        # TODO: test when type is given as string

    def is_builtin(self, type):
        # FIXME: probably some bugs in this function (not exaustive?)
        return type in [str, int, float, bool]

    def decorate_with_attr(self, cls):
        if not hasattr(cls, "__attrs_attrs__"):
            attr.s(cls)

    def get_or_create_cls_metadata(self, cls_name: str) -> ClsMetadata:
        if cls_name in self.classes:
            return self.classes[cls_name]
        else:
            cls_metadata = ClsMetadata(cls_name)
            self.classes[cls_name] = cls_metadata
            return cls_metadata

    def __len__(self):
        return self.classes.__len__()

    def __contains__(self, item):
        return self.classes.__contains__(item)

    def __getitem__(self, item) -> ClsMetadata:
        return self.classes.__getitem__(item)


class BaseRelationalDB:
    def __init__(self):
        self.db_metadata = DBMetadata()

    @optionalargs
    def register(self, cls):
        self.db_metadata.register(cls)
        return cls

    def close(self):
        pass

    def sync(self):
        pass

    def attribute(
        self,
        # attrs attribute
        type=None,
        *,
        default=attr.NOTHING,
        validator=None,
        repr=True,
        cmp=None,
        hash=None,
        init=True,
        metadata=None,
        converter=None,
        factory=None,
        kw_only=False,
        eq=None,
        order=None,
        on_setattr=None,
        # sqlalchemy attribute
        unique=None,
        ref=None,
    ):

        metadata = self.metadata(unique=unique, ref=ref)
        return attr.ib(  # noqa
            default=default,
            validator=validator,
            repr=repr,
            cmp=cmp,
            hash=hash,
            init=init,
            metadata=metadata,
            type=type,
            converter=converter,
            factory=factory,
            kw_only=kw_only,
            eq=eq,
            order=order,
            on_setattr=on_setattr,
        )

    ib = attribute

    def metadata(self, unique=None, ref=None):
        metadata = AttrMetadata(unique=unique, ref=ref)
        return {PACKAGE_NAME: metadata}


class TestRelationalDB:
    def __init__(self, db_factory):
        self.db_factory = db_factory

    def test_api(self):
        db = self.db_factory()


class RelationalListMany(collections.UserList):
    def __init__(self, ref_attribute_name: str):
        # FIXME
        self.ref_attribute_name = ref_attribute_name

    def append(self, item: typing.T) -> None:
        self.data.append(item)
        ref_attribute = getattr(item, self.ref_attribute_name)
        ref_attribute.data.append(self)



import koalak
import pytest
from koalak.utils import temp_pathname


# ------ VARIABLES -------
def test_init_variables_in_mkframework():
    framework = koalak.mkframework(variables={"age": 30})
    assert framework.variables["age"] == 30


def test_create_variable():
    framework = koalak.mkframework()
    framework.variables["name"] = "test"
    assert framework.variables["name"] == "test"


def test_update_variable():
    framework = koalak.mkframework(variables={"name": "test"})
    assert framework.variables["name"] == "test"
    framework.variables["name"] = 30
    assert framework.variables["name"] == 30


def test_not_allowed_keyword_variables():
    with pytest.raises(TypeError):
        koalak.mkframework(variables={"home": "path"})

    framework = koalak.mkframework()
    with pytest.raises(TypeError):
        framework.variables["home"] = "test"


def test_substitute_string():
    framework = koalak.mkframework()
    framework.variables["age"] = 25
    assert framework.substitute_string("$age") == "25"
    framework.variables["age"] = 26
    assert framework.substitute_string("my age is $age") == "my age is 26"


def test_default_variables_are_present():
    framework = koalak.mkframework()
    assert "userhome" in framework.variables

    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        print("variables", framework.variables)
        assert framework.variables["home"] == homepath

    # FIXME: what is the expected behaviour of 'home' variable for unamed framework


def test_remove_variable():
    framework = koalak.mkframework()
    framework.variables["age"] = 25
    assert "age" in framework.variables
    del framework.variables["age"]
    assert "age" not in framework.variables


def test_substitutable_variables():
    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        framework.variables.set("mypath", "$home/mypath", substitute=True)
        assert framework.variables["mypath"] == f"{homepath}/mypath"

    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        framework.variables.set("mypath", "$home/mypath", substitute=False)
        assert framework.variables["mypath"] == f"$home/mypath"

    with temp_pathname() as homepath:
        framework = koalak.mkframework(homepath)
        framework.variables.set("mypath", "$home/mypath")
        assert framework.variables["mypath"] == f"$home/mypath"



import pytest
from koalak import SubcommandParser


def test_run_cmd(capsys):
    main_command = SubcommandParser()

    def run_mycmd(args):
        print(1)

    mycmd_command = main_command.add_subcommand("mycmd")
    mycmd_command.function = run_mycmd

    # assert that "mycmd" parse
    main_command.parse_args(["mycmd"])

    capsys.readouterr()  # ignore the first prints
    main_command.run(["mycmd"])
    captured = capsys.readouterr()
    assert captured.out == "1\n"


def test_run_with_args(capsys):
    main_command = SubcommandParser()

    def run_mycmd(args):
        print(args.printme)

    mycmd_command = main_command.add_subcommand("mycmd")
    mycmd_command.add_argument("printme")
    mycmd_command.function = run_mycmd

    # assert that "mycmd" parse
    main_command.parse_args(["mycmd", "1"])

    with pytest.raises(SystemExit):
        main_command.parse_args(["mycmd"])

    capsys.readouterr()  # ignore the first prints

    main_command.run(["mycmd", "test"])
    captured = capsys.readouterr()
    assert captured.out == "test\n"

    main_command.run(["mycmd", "22"])
    captured = capsys.readouterr()
    assert captured.out == "22\n"


def test_run_with_many_args(capsys):
    main_command = SubcommandParser()

    def run_mycmd(args):
        print(args.printme * args.n)

    mycmd_command = main_command.add_subcommand("mycmd")
    mycmd_command.add_argument("printme")
    mycmd_command.add_argument("-n", type=int, default=1)
    mycmd_command.function = run_mycmd

    # assert that "mycmd" parse
    main_command.parse_args(["mycmd", "1"])

    # assert that "mycmd" parse
    main_command.parse_args(["mycmd", "1", "-n", "2"])

    with pytest.raises(SystemExit):
        main_command.parse_args(["mycmd"])

    capsys.readouterr()  # ignore the first prints

    main_command.run(["mycmd", "test"])
    captured = capsys.readouterr()
    assert captured.out == "test\n"

    main_command.run(["mycmd", "hey", "-n", "3"])
    captured = capsys.readouterr()
    assert captured.out == "heyheyhey\n"


def test_nested_commands(capsys):
    main_command = SubcommandParser()

    """
    main
        - version
        - var
            - ls
            - add
    """
    version_command = main_command.add_subcommand("version")
    version_command.function = lambda args: print("version")

    var_command = main_command.add_subcommand("var")

    ls_command = var_command.add_subcommand("ls")
    ls_command.function = lambda args: print("ls")

    add_command = var_command.add_subcommand("add")
    add_command.function = lambda args: print("add")

    capsys.readouterr()  # ignore the first prints
    main_command.run(["version"])
    captured = capsys.readouterr()
    assert captured.out == "version\n"

    main_command.run(["var", "ls"])
    captured = capsys.readouterr()
    assert captured.out == "ls\n"

    main_command.run(["var", "add"])
    captured = capsys.readouterr()
    assert captured.out == "add\n"


def test_nested_commands_depth_2(capsys):
    main_command = SubcommandParser("main")

    main_command.add_subcommand("a").function = lambda args: print("d1.a")
    main_command.add_subcommand("b").function = lambda args: print("d1.b")

    depth_command = main_command.add_subcommand("depth")
    depth_command.add_subcommand("a").function = lambda args: print("d2.a")
    depth_command.add_subcommand("b").function = lambda args: print("d2.b")

    depth3_command = depth_command.add_subcommand("depth")
    depth3_command.add_subcommand("a").function = lambda args: print("d3.a")
    depth3_command.add_subcommand("b").function = lambda args: print("d3.b")

    capsys.readouterr()  # ignore the first prints
    main_command.run(["a"])
    captured = capsys.readouterr()
    assert captured.out == "d1.a\n"

    capsys.readouterr()  # ignore the first prints
    main_command.run(["b"])
    captured = capsys.readouterr()
    assert captured.out == "d1.b\n"

    with pytest.raises(SystemExit):
        main_command.run([])

    # depth2
    capsys.readouterr()  # ignore the first prints
    main_command.run(["depth", "a"])
    captured = capsys.readouterr()
    assert captured.out == "d2.a\n"

    capsys.readouterr()  # ignore the first prints
    main_command.run(["depth", "b"])
    captured = capsys.readouterr()
    assert captured.out == "d2.b\n"

    with pytest.raises(SystemExit):
        main_command.run(["depth"])

    # depth3
    capsys.readouterr()  # ignore the first prints
    main_command.run(["depth", "depth", "a"])
    captured = capsys.readouterr()
    assert captured.out == "d3.a\n"

    capsys.readouterr()  # ignore the first prints
    main_command.run(["depth", "depth", "b"])
    captured = capsys.readouterr()
    assert captured.out == "d3.b\n"

    with pytest.raises(SystemExit):
        main_command.run(["depth", "depth"])

    # None existing commands/args
    with pytest.raises(SystemExit):
        main_command.run(["a", "dontexist"])

    with pytest.raises(SystemExit):
        main_command.run(["depth", "dontexist"])

    with pytest.raises(SystemExit):
        main_command.run(["depth", "depth", "dontexist"])

    with pytest.raises(SystemExit):
        main_command.run(["a", "-x"])


def test_nested_commands_fullname(capsys):
    main_command = SubcommandParser("main")
    assert main_command.name == "main"
    assert main_command.fullname == "main"

    a_main_command = main_command.add_subcommand("a")
    assert a_main_command.name == "a"
    assert a_main_command.fullname == "main.a"

    b_main_command = main_command.add_subcommand("b")
    assert b_main_command.name == "b"
    assert b_main_command.fullname == "main.b"

    # check that main name didn't change
    assert main_command.name == "main"
    assert main_command.fullname == "main"

    depth_command = main_command.add_subcommand("depth")
    assert depth_command.name == "depth"
    assert depth_command.fullname == "main.depth"

    main_depth_a_cmd = depth_command.add_subcommand("a")
    assert main_depth_a_cmd.name == "a"
    assert main_depth_a_cmd.fullname == "main.depth.a"

    main_depth_b_cmd = depth_command.add_subcommand("b")
    assert main_depth_b_cmd.name == "b"
    assert main_depth_b_cmd.fullname == "main.depth.b"

    main_depth_depth_cmd = depth_command.add_subcommand("depth")
    assert main_depth_depth_cmd.name == "depth"
    assert main_depth_depth_cmd.fullname == "main.depth.depth"

    main_depth_depth_a_cmd = main_depth_depth_cmd.add_subcommand("a")
    assert main_depth_depth_a_cmd.name == "a"
    assert main_depth_depth_a_cmd.fullname == "main.depth.depth.a"

    main_depth_depth_b_cmd = main_depth_depth_cmd.add_subcommand("b")
    assert main_depth_depth_b_cmd.name == "b"
    assert main_depth_depth_b_cmd.fullname == "main.depth.depth.b"


# TODO: test without function linked

# Test errors
def test_subcmdparser_without_subcommands(capsys):
    # Without subcommands should have a function or error
    main_command = SubcommandParser(description="catchme")

    with pytest.raises(ValueError):
        main_command.parse_args([])

    # if main_command have function no error is raised
    main_command.function = lambda args: print("hello")
    main_command.parse_args([])


def test_subcmdparser_subcmd_without_function():
    main_command = SubcommandParser()
    subcmd_command = main_command.add_subcommand("subcmd")

    with pytest.raises(ValueError):
        main_command.run([])


def test_run_non_existing_cmd(capsys):
    # run non-existing command print help and exit
    main_command = SubcommandParser()
    main_command.add_subcommand("mycmd").function = lambda args: print("hello")

    with pytest.raises(SystemExit):
        capsys.readouterr()  # ignore last entries
        main_command.run(["dontexist"])

    # help was printed
    assert "invalid choice" in capsys.readouterr().err.lower()


def test_run_attribute_not_callable():
    main_command = SubcommandParser()

    main_command.add_subcommand("x").function = "String and not callable"

    with pytest.raises(TypeError):
        main_command.parse_args([])


def test_subcmd_already_exists():
    main_command = SubcommandParser()

    main_command.add_subcommand("x")

    with pytest.raises(KeyError):
        main_command.add_subcommand("x")


def test_help_basic(capsys):
    main_command = SubcommandParser("main")
    a_cmd = main_command.add_subcommand("a", description="do nothing")
    capsys.readouterr()

    main_command.print_help()
    expected_help = """Usage: main [-h] <subcommand>

Commands:

    a    do nothing"""
    # print()
    # print(expected_help)
    assert capsys.readouterr().out.strip() == expected_help


def test_help_3_subcmd(capsys):
    main_command = SubcommandParser("main")
    a_cmd = main_command.add_subcommand("a", description="do nothing")
    a_cmd = main_command.add_subcommand("b", description="yes")
    a_cmd = main_command.add_subcommand("cc", description="no")
    capsys.readouterr()

    main_command.print_help()
    # add more spaces because cc have length of 2
    expected_help = """Usage: main [-h] <subcommand>

Commands:

    a     do nothing
    b     yes
    cc    no"""
    # print()
    # print(expected_help)
    assert capsys.readouterr().out.strip() == expected_help


def test_help_hide(capsys):
    main_command = SubcommandParser("main")
    a_cmd = main_command.add_subcommand("a", description="do nothing")
    a_cmd = main_command.add_subcommand("b", description="yes", hide=True)
    a_cmd = main_command.add_subcommand("cc", description="no")
    capsys.readouterr()

    main_command.print_help()
    # add more spaces because cc have length of 2
    expected_help = """Usage: main [-h] <subcommand>

Commands:

    a     do nothing
    cc    no"""
    # print()
    # print(expected_help)
    assert capsys.readouterr().out.strip() == expected_help


def test_help_adding_groups(capsys):
    main_command = SubcommandParser("main")

    main_command.add_group(
        "first", title="First commands", description="important commands"
    )
    main_command.add_group(
        "second", title="Second commands", description="useless commands"
    )

    main_command.add_subcommand("a", description="do nothing", group="first")
    main_command.add_subcommand("bbb", description="yes", hide=True)
    main_command.add_subcommand("cc", description="no", group="second")
    main_command.add_subcommand("d", description="yesno", group="first")

    capsys.readouterr()

    main_command.print_help()
    # add more spaces because cc have length of 2
    expected_help = """Usage: main [-h] <subcommand>

First commands: important commands

    a     do nothing
    d     yesno

Second commands: useless commands

    cc    no
"""
    # print()
    # print(expected_help)
    assert capsys.readouterr().out.strip() == expected_help.strip()
